# -*- python -*-
# $RCSfile: output.py,v $
# $Revision: 1.30 $
# $Author: langer $
# $Date: 2004/11/24 16:44:10 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov.

from SWIG.common import coord
from SWIG.common import timestamp
from SWIG.common import switchboard
from SWIG.common import ooferror
from common import debug
from common import primitives
from common import utils
from common.IO import parameter
from types import *
import math, string

class InputMismatch(parameter.ParameterMismatch): pass

# Derive some types from parameter.Parameter to provide type and
# computability checking.

class Input(parameter.ParameterBase):
    """
    An Output's Inputs are other Outputs that are connected to it.  An
    Output can be computed only if all of its Inputs can be computed.
    The type of an Input is the output type of its Output. Third base.
    """
    def __init__(self, name, types, value=None, default=None, tip=None):
        if isinstance(types,parameter.TypeChecker):
            self.checker = types
        else:
            if type(types) == ListType or type(types) == TupleType:
                self.checker = parameter.TypeChecker(*types)
            else:
                self.checker = parameter.TypeChecker(types)
        parameter.ParameterBase.__init__(self, name, value, default, tip)
    def set(self, value):               # value is another Output
        if value != None:
            try:
                self.checker(value.oType)
            except TypeError, msg:
                raise InputMismatch(msg)
        self.__dict__['value'] = value
        self.timestamp.increment()

    
    def __repr__(self):
        # TypeChecker.__repr__ returns a list of types, so we use
        # self.checker here, instead of the 'types' constructor
        # argument (which isn't stored anywhere other than in the
        # TypeChecker anyway).
        if self.value is None:
            return '%s("%s", %s)' % (self.__class__.__name__,
                                     self.name, `self.checker`)
        else:
            return '%s("%s", %s, %s)' % (self.__class__.__name__,
                                         self.name, `self.checker`,
                                         `self.value`)

    def clone(self):
        if self.value:
            value = self.value.clone()
        else: 
            value = None
        return Input(self.name, self.checker, value=value)

    def evaluate(self, mesh, elements, coords):
        return self.value.evaluate(mesh, elements, coords)

##    def getParameter(self, name):
##        return self.value.getParameter(name)


#########################################################
    
class Output:
    """
    An instance of Output will act as an output 'class'.  The __init__
    will register the Output instance so that it can be copied. An
    Output can be generated that is computable, or it may be lacking
    some of its defined inputs or parameters. An Output that has all
    required inputs defined, all of which are themselves 'computable'
    will itself be computable.

    I think this paragraph is obsolete:
    Allowed output types are any of the basic types in the 'types'
    module. ListType outputs may be assumed to be 2-dimensional.
    A more general mechanism for matching types may be implemented
    if is seems useful. As a special case, 'None' may be used for
    the output type to indicate that the output could return
    anything. Likewise, None can indicate that any input type
    is acceptable.

    parameters is a list of Parameter objects.

    If 'parent', then this is a copy of another Output, so it should
    not be registered.

    Outputs marked as 'prototypes' are used to determine which Output
    parameters need to be set by the UI.  Ie, parameters which are
    unset in the prototype need to be settable in the GUI.  In some
    sense, a prototype Output is akin to a class, whereas a clone of a
    prototype is akin to an instance.  Perhaps this analogy should be
    pursued in detail and made explicit.

    """

    # All Outputs with parent=None are automatically made into
    # prototypes at registration time, independently of the
    # "prototype" setting here.  Not all prototypes are parents,
    # however.
    
    def __init__(self, func, name, oType=None, inputs=None,
                 parameters=None, tip=None, discussion=None,
                 parent=None, prototype=1):

        self._name = name
        self.tip = tip
        self.discussion = discussion
        self.parent = parent
        self.prototype = prototype

        self.oType = None
        if isinstance(oType, parameter.TypeChecker):
            self.oType = oType
        elif type(oType) in (ListType, TupleType):
            self.oType = parameter.TypeChecker(*oType)
        else:
            self.oType = parameter.TypeChecker(oType)

        if not callable(func):
            raise TypeError("func must be a callable")
        self.func = func

        self.inputs = []
        if inputs:
            for i in utils.listify(inputs):
                if isinstance(i,Input):
                    self.inputs.append(i)
                elif isinstance(i,parameter.TypeChecker):
                    self.inputs.append(Input("",i))
                else:
                    self.inputs.append(Input("",i,i))

        self.parameters = []
        if parameters:
            for p in utils.listify(parameters):
                if isinstance(p, parameter.ParameterBase):
                    self.parameters.append(p)
                else:
                    
                    raise ooferror.ErrPyProgrammingError(
                        "Expected a subclass of ParameterBase.")

        self.timestamp = timestamp.TimeStamp() # creation or modification time

    def name(self):
        return self._name

    def clone(self, func=None, name=None, oType=None, tip=None, discussion=None,
              inputs=[], parameters=[]):
        """
        'self' is (generally) a template. Recursively duplicates self
        and all of self's inputs and parameters. Note that complicated
        parameters (like class instances) will wind up being shared
        between all clones. Most characteristics of self can be
        overridden in the clone.
        """
        if not func:
            func = self.func
        if not name:
            name = self._name
        if not oType:
            oType = self.oType
        if not inputs and self.inputs:
            inputs = [i.clone() for i in self.inputs]
        if not parameters and self.parameters:
            parameters = [p.clone() for p in self.parameters]
        if not tip:
            tip = self.tip
        if not discussion:
            discussion = self.discussion
        return Output(func=func, name=name, oType=oType,
                      inputs=inputs, parameters=parameters, tip=tip,
                      discussion=discussion,
                      parent=self, prototype=0)

    def __call__(self, **kwargs):
        # In the spirit of making Outputs, which are instances, act
        # more or less like classes, here we define the __call__
        # function to act as if it's creating an instance with
        # specified parameters.
        clone = self.clone()
        for key, value in kwargs.items():
            clone.plugParameter(which=key, value=value)
        return clone
            
        
    def plugInput(self, input, which=0, absolute=0, deep=0):
        """
        Sticks the given input into the list of inputs in the open
        position given by 'which'. If 'absolute', then both filled
        and unfilled inputs will be counted, otherwise only unfilled
        inputs will be counted. If 'deep', then the count will
        recurse through already plugged inputs. Use of both 'absolute'
        and 'deep' is probably not recommended, since the results
        may be surprising. The return value is only used for recursing.
        """
        if not self.inputs:
            raise ValueError("Output takes no inputs")

        self.timestamp.increment()
        
        if type(which) == StringType: # Set input by name
            # If the name is of the form 'part1/part2/...',
            # then a sub-input is being addressed. If the name
            # is of the form 'input/', then force replacement
            # of input in self. Otherwise, search recursively for
            # an open input having the correct name.
            slash = string.find(which,"/")
            if slash > -1:
                name = which[:slash]
                which = which[slash+1:]
                for i in self.inputs:
                    if i.name == name:
                        if which == "": # / was at the end of the string
                            i.value = input
                        else:
                            i.value.plugInput(input,which)
                        return
                raise ValueError("No such input: %s" % which )
            else: # No /
                # First see if we've an empty input having that name
                for i in self.inputs:
                    if i.name == which:
                        i.value = input
                        return
                # So much for that. Now try sticking it into each 
                # input in turn to see whether it matches.
                for i in self.inputs:
                    try:
                        if not i.value is None:
                            i.value.plugInput(input,which)
                            return
                    except ValueError:
                        pass
                raise ValueError("No such input: %s" % which)
        else: # Numerical 'which'
            for idx in range(len(self.inputs)):
                if absolute or not self.inputs[idx]:
                    which -= 1
                if which < 0:
                    self.setInput(input,idx)
                    return
                # All counting from left to right, even going into
                # sub-inputs
                if deep and self.inputs[i]:
                    which = self.inputs[i].plugInput(input,which,absolute,deep)
                if which < 0:
                    break
            return which

    def plugParameter(self, value=None, default=None,
                      which=0, absolute=0, deep=0, rename=None, tip=None):
        """
        Sticks the given parameter into the list of parameters in the open
        position given by 'which'. If 'absolute', then both filled
        and unfilled parameters will be counted, otherwise only unfilled
        parameters will be counted. If 'deep', then the count will
        recurse through already plugged parameters. Use of both 'absolute'
        and 'deep' is probably not recommended, since the results
        may be surprising. The return value is only used for recursing.
        """
        if not self.parameters:
            raise ValueError("Output takes no parameters")

        self.timestamp.increment()
        
        if type(which) == StringType: # Set parameter by name
            # If the name is of the form 'part1/part2/...',
            # then a parameter in an input is being addressed. If the name
            # is of the form 'parameter/', then force replacement
            # of parameter in self. Otherwise, search recursively for
            # an open parameter having the correct name.
            slash = string.find(which,"/")
            if slash > -1:
                name = which[:slash]
                which = which[slash+1:]
                if which == "": # / was at the end of the string
                    for p in self.parameters:
                        if p.name == name:
                            if value is not None:
                                p.value = value
                            if default is not None:
                                p.default = default
                            if rename is not None:
                                p.name = rename
                            if tip is not None:
                                p.tip = tip
                            return
                    raise ValueError("No such parameter: %s" % which )
                else: # Wasn't just a terminating /
                    for i in self.inputs:
                        if i.name == name:
                            i.value.plugParameter(value,default,which,tip=tip)
                        return
                    raise ValueError("No such input: %s" % name)
            else: # No /
                # First see if we've an empty parameter having that name
                for p in self.parameters:
                    if p.name == which:
                        if value is not None:
                            p.value = value
                        if default is not None:
                            p.default = default
                        if rename is not None:
                            p.name = rename
                        if tip is not None:
                            p.tip = tip
                        return
                # So much for that. Now try sticking it into each 
                # input in turn to see whether it matches.
                for input in self.inputs:
                    try:
                        input.value.plugParameter(value,default,which)
                        return
                    except ValueError:
                        pass
                raise ValueError("No such parameter: %s"%which)
        else: # Numerical 'which'
            raise "Setting parameters by number is not implemented!"
##            # left-to-right doesn't mean much when you're talking
##            # about filling parameters in inputs, so just count all
##            # 'our' parameters first, then move on to the inputs from
##            # left-to-right.
##            for idx in range(len(self.parameters)):
##                if absolute or not self.parameters[idx]:
##                    which -= 1
##                if which < 0:
##                    self.setParameter(value,default,idx)
##                    return
##            if deep:
##                for input in self.inputs:
##                    which = input.value.plugParameter(
##                        value,default,which,absolute,deep)
##                    if which < 0:
##                        return
##            return which

##    def getParameter(self, name):
##        for p in self.parameters:
##            if p.name == name:
##                return p
##        for input in self.inputs:
##            p = input.getParameter(name)
##            if p is not None:
##                return p

    def listAllParameters(self):
        plist = []
        if self.inputs:
            for input in self.inputs:
                plist += input.value.listAllParameters()
        plist += self.parameters[:]
        return plist

    def listAllParametersHierarchically(self):
        phier = []
        if self.inputs:
            for input in self.inputs:
                ihier = input.value.listAllParametersHierarchically()
                if len(ihier) == 1:
                    phier += ihier
                elif len(ihier) > 1:
                    phier.append(ihier)
        if len(self.parameters) == 1:
            phier += self.parameters
        elif len(self.parameters) > 1:
            phier.append(self.parameters[:])
        return phier

    def getPrototype(self):
        # Find the nearest ancestor that's a prototype
        if self.prototype:
            return self
        if self.parent:
            return self.parent.getPrototype()
        return None

    def getPath(self):
        return self.getPrototype()._path()
    def _path(self):                    # only called on prototypes
        for tree in outputTrees:
            try:
                return tree.objpath(self)
            except KeyError:
                pass
        raise ooferror.ErrPyProgrammingError("Can't find path to Output object")
    
    def getSettableParams(self):
        # Settable parameters are parameters which don't have values
        # in the prototype.  This returns clones of the parameters, so
        # to actually set them in this object, explicitly call
        # plugParameter.  It's done this way so that the *unset*
        # parameters in prototypes stay that way.
        params = self.listAllParameters()
        protoparams = self.getPrototype().listAllParameters()
        unsetparams = [p.clone() for (p, pp) in zip(params, protoparams)
                       if pp.value is None]
        for p in unsetparams:
            if p.value is None:
                p.value = p.default
        return unsetparams

    def getSettableParamsHierarchically(self):
        params = self.listAllParametersHierarchically()
        protoparams = self.getPrototype().listAllParametersHierarchically()
        return _weed_and_copy(params, protoparams)
        

    def setDefaultParams(self):
        protoparams = self.getPrototype().listAllParameters()
        params = self.listAllParameters()
        for p, pp in zip(params, protoparams):
            if pp.value is None:
                pp.default = p.value
        

    def incomputable(self, mesh):
        ## Checks to see if we are complete: all required inputs and
        ## parameters are defined, and all inputs are themselves
        ## computable. Actually, checks for incompleteness, because
        ## it's easier...
        for i in self.inputs:
            if i.value == None:
                return ("input unset", self._name, i.name)
            incomp = i.value.incomputable(mesh)
            if incomp:
                return incomp 
        for p in self.parameters:
            incomp = p.incomputable(mesh)
            if incomp:
                return incomp
        return 0

        
    def evaluate(self,mesh,elements,coords):
        """
        If the list of coords lies in multiple elements, they need to
        be subdivided into a list of lists that match up with the list
        of elements. For 'elements', it would be possible to also read
        'edges'.
        """
        assert len(coords) == len(elements)
        inputResults = [input.evaluate(mesh,elements,coords)
                        for input in self.inputs]
        return apply(self.func,
                     inputResults,
                     {"mesh":mesh,
                      "elements":elements,
                      "coords":coords,
                      "parameters":self.parameters})

    def __repr__(self):
        return "getOutput(%s)" % self.short_repr()

    def short_repr(self):
        params = self.getSettableParams()
        args = ["'%s'" % self.getPath()] + ['%s=%s' % (p.name, `p.value`)
                                       for p in params]
        return "%s" % string.join(args, ',')

    def getTimeStamp(self):
        # Get the latest timestamp that affects this output.
        if self.inputs:
            inputtime = max([input.getTimeStamp() for input in self.inputs])
        else:
            inputtime = timestamp.timeZero
        if self.parameters:
            paramtime = max([param.getTimeStamp() for param in self.parameters])
        else:
            paramtime = timestamp.timeZero
        return max(inputtime, paramtime, self.timestamp)


################################################################

# Outputs are stored in LabelTrees according to their oType.  These
# trees are used to construct the GUI widgets from which the user
# chooses an Output.  Outputs are put into the trees by the functions
# definePositionOutput and defineScalarOutput.  When an Output is put
# into a tree, it's marked as a prototype, and can then be used to
# create specific instances. When the UI is searching an Output's
# ancestry to find which Parameters should be set by the GUI, it stops
# when it finds a prototype Output.

from common import labeltree

class OutputTree(labeltree.LabelTree):
    def __setitem__(self, label, out):
        labeltree.LabelTree.__setitem__(self, label, out)
        out.prototype = 1

positionOutputs = OutputTree()
scalarOutputs = OutputTree()
aggregateOutputs = OutputTree()

outputTrees = (positionOutputs, scalarOutputs, aggregateOutputs)

def definePositionOutput(path, output):
    positionOutputs[path] = output

# A Parameter whose value is an Output whose value is a Position.
class PositionOutputParameter(parameter.ParameterBase):
    oktypes = parameter.TypeChecker(primitives.Point, coord.Coord)
    def checker(self, x):
        if not isinstance(x, Output):
            parameter.raiseTypeError(type(x), 'Position Output')
        PositionOutputParameter.oktypes(x.oType)
    def valueDesc(self):
        return "An <link linkend='Section:Output:Position'><classname>Output</classname></link> object whose value is a position in the xy plane."
        
def defineScalarOutput(path, output):
    scalarOutputs[path] = output

# A Parameter whose value is an Output whose value is a scalar.
class ScalarOutputParameter(parameter.ParameterBase):
    oktypes = parameter.TypeChecker(IntType, FloatType)
    def checker(self, x):
        if not isinstance(x, Output):
            parameter.raiseTypeError(type(x), 'Scalar Output')
        ScalarOutputParameter.oktypes(x.oType)
    def valueDesc(self):
        return "An <link linkend='Section:Output:Scalar'><classname>Output</classname></link> object whose value is a real number."

# AggregateOutputs are outputs whose values are possibly complex
# objects, but which should be displayed together.  For example, all
# the components of a tensor flux are and AggregateOutput.

def defineAggregateOutput(path, output):
    aggregateOutputs[path] = output

class AggregateOutputParameter(parameter.ParameterBase):
    def checker(self, x):
        if not isinstance(x, Output):
            parameter.raiseTypeError(type(x), 'Output')
    def valueDesc(self):
        return "An <link linkend='Section:Output:Aggregate'><classname>AggregateOutput</classname></link> object."

def getOutput(name, **kwargs):
    # Look for the prototype output
    try:
        proto = scalarOutputs[name]
    except KeyError:
        try:
            proto = positionOutputs[name]
        except KeyError:
            raise KeyError("Unknown Output: %s" % name)
    return proto.object(**kwargs)  # clones an Output

utils.OOFdefine('getOutput', getOutput)

##################

## utility function used by Output.getSettableParamsHierarchically().
## params and protoparams are lists of lists of lists ... of
## Parameters.  The lists are nested to an unknown and varying depth.
## Return a hierearchical list with the same structure containing
## clones of the parameters in the params list, but containing only
## those parameters that have value==None in the protoparams list.  If
## a parameter has value==None in the params list, set its value to
## the default value of the corresponding protoparam.  Finally, remove
## empty sublists.

def _weed_and_copy(params, protoparams):
    result = []
    for p, pp in zip(params, protoparams):
        if isinstance(p, parameter.ParameterBase):
            if pp.value is None:
                bozo = p.clone()
                if bozo.value is None:
                    bozo.value = pp.default
                result.append(bozo)
        else:                           # p and pp should be lists
            sublist = _weed_and_copy(p, pp)
            if sublist:
                result.append(sublist)
    return result
