# -*- python -*-
# $RCSfile: parameter.py,v $
# $Revision: 1.91 $
# $Author: langer $
# $Date: 2004/12/14 22:56:06 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 



# Parameter objects are named, type-aware containers, and are known to
# the user interface.  There are subclasses of ParameterBase for all
# the basic types, and a number of composite but not object types
# (i.e. TupleOfInts parameter).  Although the functionality of the
# classes is similar, having separate classes allows for custom
# widgets to be attached to the class for the GUI, as well as for
# customized binary representations of the parameter values.

# For RegsiteredClass objects, the Parameter function is provided,
# which will automatically look up the registry of the passed-in type
# class or instance, and will return an object of the appropriate
# sort, either RegisteredParameter or ConvertibleRegisteredParameter.

# Parameter objects are used extensively in the user interface.  The
# name of the parameter will very likely be exposed to the user, and
# so should make sense in that context -- sometimes this may result in
# confusion in other contexts, e.g. there are parameters whose names
# are "skeleton", but which return strings, rather than skeleton
# objects -- this is because the user is in fact being prompted for
# the *name* of a skeleton.

# Most parameters have simple constructors which take a name, an
# initial value, a default value, and various class-specific keyword
# arguments, one of which should be a string-valued thing called a
# "tip" -- this is used by the GUI to put helpful tooltips up for the
# user.

from SWIG.common import ooferror
from SWIG.common import timestamp
from common import debug
from common import registeredclass
from common import utils
from common.IO import automatic
from common.IO.typename import typename
from types import *
import string
import struct

ErrPyProgrammingError = ooferror.ErrPyProgrammingError

structIntFmt = '>i'
structIntSize = struct.calcsize(structIntFmt)

def raiseTypeError(got, expected):
    raise TypeError('Bad type!\nExpected %s\nGot %s' % (`expected`, `got`))

class TypeChecker:
    def __init__(self, *types):
        self.types = []
        for t in types:
            ttype = type(t)
            if ttype == ClassType or ttype == TypeType:
                self.types.append(t)
            elif ttype == InstanceType:
                self.types.append(t.__class__)
            else:
                self.types.append(ttype)
    def __call__(self, x):               # raise TypeError if x is not allowed
        if x == None: # None is a special case that matches everything
            return
        if isinstance(x, TypeChecker):
            # Check that all types allowed by x are allowed by self.
            for t in x.types:
                try:
                    self(t)
                except TypeError:
                    raiseTypeError(x.types, self.types)
            return
        if type(x) == InstanceType:
            xtype = x.__class__
        else:
            xtype = type(x)
        if xtype == TypeType:
            for c in self.types:
                if c == x:
                    return
        elif xtype == ClassType:
            for c in self.types:
                if issubclass(x,c):
                    return
        else:
            for c in self.types:
                if isinstance(x,c):
                    return
        raiseTypeError(xtype, self.types)
    def __repr__(self):
        if len(self.types)==1:
            return typename(self.types[0])
        return '[' + string.join([typename(t) for t in self.types], ',') + ']'
    def __len__(self):
        return len(self.types)
    def __getitem__(self, i):
        return self.types[i]

class IndexableTypeChecker(TypeChecker):
    def __call__(self, x):
        return utils.isIndexable(x)
    def __repr__(self):
        return 'IndexableTypeChecker()'

##class CallableTypeChecker(TypeChecker):
##    def __call__(self, x):
##        return callable(x)
##    def __repr__(self):
##        return 'CallableTypeChecker()'

class ParameterMismatch(TypeError): pass

## Base class for Parameters.  The derived classes must provide a
## 'checker' method (or callable instance) which takes a parameter
## value as an argument and raises TypeError if the value is invalid.
## 'checker' is called by the base class __init__, so whatever data it
## requires needs to be present before the derived class's __init__
## calls the base class __init__.  The class must also provide a a
## __repr__ and possibly a clone() (if the supplied clone() is
## insufficient).

## The 'default' argument to the constructor is used in cases where
## it's necessary to distinguish between variables that are set and
## those that are unset (ie have value==None) but it's nonetheless
## necessary to display a value for the unset variables.  The
## displayed value can be retrieved from 'default'.  For an example,
## see the Output class in common/IO/output.py.

class ParameterBase:
    def __init__(self, name, value=None, default=None, tip=None):
        self.name = name
        self.tip = tip
        self.timestamp = timestamp.TimeStamp()
        self.group = None
        self.default = default          # used by Output
        # Internally, don't use "self.value = x" unless you really
        # want to call __setattr__!
        if value is not None:
            self.set(value)
        else:
            self.__dict__['value'] = None
        
    def set(self, value):
        if value is not None:
            try:
                self.checker(value)
            except TypeError, msg:
                raise ParameterMismatch(str(msg) + '\nfor Parameter: '
                                        + self.name)
        self.__dict__['value'] = self.converter(value)
        self.timestamp.increment()

    # Subclasses can override converter() to ensure that stored values
    # have the correct form.
    def converter(self, value):
        return value

    def __repr__(self):
        return "%s(name='%s', value=%s, default=%s)" % \
               (self.__class__.__name__, self.name, self.value, self.default)

    def getTimeStamp(self):             # When was this parameter changed?
        return self.timestamp

    def incomputable(self, context):    # used by Output
        if self.value is None:
            return "Parameter '%s' has no assigned value." % self.name

    # Intercept the assignment operator to check the type of 'value'.
    def __setattr__(self, attr, value):
        if attr == 'value':
            self.set(value)
        else:
            self.__dict__[attr] = value

    def set_group(self, group):
        # Called by ParameterGroup when this Parameter is added to a group.
        self.group = group
            
    def clone(self):
        # Override this if the derived class has any local data!
        return self.__class__(self.name, self.value, self.default, self.tip)

    def makeWidget(self, *args, **kwargs):
        # This function must be overridden in derived classes in GUI mode.
        raise ooferror.ErrPyProgrammingError(
            'No widget for %s!' % self.__class__.__name__)


    def binaryRepr(self, datafile, value):
        raise ooferror.ErrPyProgrammingError(
            "Somebody forgot to define %s.binaryRepr()!"
            % self.__class__.__name__)

    def binaryRead(self, parser):
        raise ooferror.ErrPyProgrammingError(
            "Somebody forgot to define %s.binaryRead()!" 
            % self.__class__.__name__)

    # The following two functions are used for documenting the user interface.
    def classRepr(self):
        # Returns the name of the type of parameter.  For most
        # parameters, this is just the class name, but for some (such
        # as RegisteredParameters) the class name isn't sufficient.
        return self.__class__.__name__
    def valueRepr(self):
        # Returns a string representing the legal values for the
        # parameter.  If the classRepr is sufficient, this doesn't
        # need to return anything.
        pass
    def valueDesc(self):
        # valueDesc returns a string that describes the allowable
        # values for the parameter.  It's used in the automatically
        # generated xml documentation.  It can contain any docbook
        # markup allowed within a <para> element.
        raise "Someone forgot to define %s.valueDesc" % self.__class__.__name__

##########

# Special Parameter subclasses.  They have specialized widgets and
# BinaryReprs.

class BooleanParameter(ParameterBase):
    def __init__(self, name, value=None, default=0, tip=None):
        ParameterBase.__init__(self, name, value, default, tip)
    def checker(self, x):
        # Comparing to 0 and 1 works for both Booleans and Ints, so
        # it's ok even for versions of Python that don't have
        # types.BooleanType.
        if x!=0 and x!=1:
            raiseTypeError(type(x), "0 or 1")
    def set(self, value):
        if value is not None:
            if not (value==0 or value==1):
                raise ParameterMismatch('Got ' + `value` + ' for Parameter '
                                        + self.name)
        self.__dict__['value'] = value
        self.timestamp.increment()
    def binaryRepr(self, datafile, value):
        if value:
            return "1"
        return "0"
    def binaryRead(self, parser):
        v = parser.getBytes(1)
        return (v == "1")
    def valueDesc(self):
        return "Boolean, 0 (false) or 1 (true)."

class _RangeParameter(ParameterBase):
    def __init__(self, name, range, types, value, default, tip):
        self.range = range
        self.types = types
        ParameterBase.__init__(self, name, value, default, tip)
    def clone(self):
        return self.__class__(self.name, self.range, self.value, self.default,
                              self.tip)
    def set(self, value):
        if value is not None:
            if type(value) not in self.types:
                raise ParameterMismatch('Got ' + `value` + ' for Parameter '
                                        + self.name)
            if self.range[0] <= value <= self.range[1]:
                self.__dict__['value'] = value
                self.timestamp.increment()
            else:
                raise ValueError('Parameter value out of range for '
                                 + self.name)
    def classRepr(self):
        return "%s%s" % (self.__class__.__name__, self.range)

class IntRangeParameter(_RangeParameter):
    def __init__(self, name, range, value=None, default=None, tip=None):
        _RangeParameter.__init__(self, name, range, [IntType], value, default,
                                 tip)
    def binaryRepr(self, datafile, value):
        return struct.pack(structIntFmt, value)
    def binaryRead(self, parser):
        b = parser.getBytes(structIntSize)
        (val,) = struct.unpack(structIntSize, b)
        return val
    def valueDesc(self):
        return "An integer in the range [%d, %d]." % \
               (self.range[0], self.range[1])

class FloatRangeParameter(_RangeParameter):
    structfmt = ">d"
    structsize = struct.calcsize(structfmt)
    def __init__(self, name, range, value=None, default=None, tip=None):
        _RangeParameter.__init__(self, name, range, [FloatType, IntType],
                                 value, default, tip)
    def binaryRepr(self, datafile, value):
        return struct.pack(FloatRangeParameter.structfmt, value)
    def binaryRead(self, parser):
        b = parser.getBytes(FloatRangeParameter.structsize)
        (val,) = struct.unpack(FloatRangeParameter.structfmt, b)
        return val
    def valueDesc(self):
        return "A real number in the range [%d, %d]." % \
               (self.range[0], self.range[1])
        
# Parameter class that can take a float or an integer,
# or the special value "automatic". 
class AutoNumericParameter(ParameterBase):
    def __init__(self, name, value=None, default=None, tip=None):
        self.checker = TypeChecker(FloatType, IntType, automatic.automatic)
        ParameterBase.__init__(self, name,
                               value=value, default=default, tip=tip)
    def valueRepr(self):
        return "Int, Float, or automatic"
    def valueDesc(self):
        return "An integer, a real number, or the string 'automatic'."


class AutoIntParameter(ParameterBase):
    def __init__(self, name, value=None, default=None, tip=None):
        self.checker = TypeChecker(IntType, automatic.automatic)
        ParameterBase.__init__(self, name, value=value, default=default,
                               tip=tip)
    def valueRepr(self):
        return "Int or automatic"
    def valueDesc(self):
        return "An integer, or the string 'automatic'."
    

# Parameter type for an integer number of values or a list of specific
# float values -- used for the set of levels in the contour maps, but
# possibly useful elsewhere.  This used to accept the value
# "automatic", but the code to handle it wasn't implemented.

class ValueSetParameter(ParameterBase):
    def __init__(self, name, value=None, default=None, tip=None):
        self.checker = TypeChecker(IntType, ListType, TupleType)
        ParameterBase.__init__(self, name,
                               value=value, default=default, tip=tip)
    def converter(self, value):
        if type(value) is ListType:
            return tuple(value)
        return value
    def valueRepr(self):
        return "Int or List of numbers"
    def valueDesc(self):
        return "An integer, or a list of numbers."



# Special parameter type for strings.  Making a subclass allows for a
# widget that doesn't require the quotes.  Note that using this
# parameter means the user can't just type in the name of a variable
# containing a string.

class StringParameter(ParameterBase):
    def __init__(self, name, value=None, default="", tip=None):
        ParameterBase.__init__(self, name, value=value, default=default,
                               tip=tip)
    def checker(self, x):
        if x is not None and type(x) is not StringType:
            raiseTypeError(type(x), StringType)
    def binaryRepr(self, datafile, value):
        length = len(value)
        return struct.pack(structIntFmt, length) + value
    def binaryRead(self, parser):
        b = parser.getBytes(structIntSize)
        (length,) = struct.unpack(structIntFmt, b)
        return parser.getBytes(length)
    def valueDesc(self):
        return "A character string."

class ListOfStringsParameter(ParameterBase):
    def __init__(self, name, value=None, default=[], tip=None):
        ParameterBase.__init__(self, name, value, default, tip)
    def checker(self, x):
        if type(x) is not ListType:
            raiseTypeError(type(x), "list of strings")
        for s in x:
            if type(s) is not StringType:
                raiseTypeError("list of %s" % typename(type(s)),
                               "list of strings")
    def binaryRepr(self, datafile, value):
        lengthstr = struct.pack(structIntFmt, len(value))
        strings = [lengthstr] + [struct.pack(structIntFmt, len(s)) + s
                                 for s in value]
        return string.join(strings, '')
    def binaryRead(self, parser):
        b = parser.getBytes(structIntSize)
        (length,) = struct.unpack(structIntFmt, b)
        strings = []
        for i in range(length):
            b = parser.getBytes(structIntSize)
            (strlen,) = struct.unpack(structIntFmt, b)
            strings.append(parser.getBytes(strlen))
        return strings
    def valueDesc(self):
        return "A list of character strings."

class FloatParameter(ParameterBase):
    def __init__(self, name, value=None, default=0.0, tip=None):
        ParameterBase.__init__(self, name, value=value, default=default,
                               tip=tip)
    def checker(self, x):
        tipe = type(x)
        if not (tipe is FloatType or tipe is IntType):
            raiseTypeError(type(x), FloatType)
    structfmt = '>d'
    structsize = struct.calcsize(structfmt)
    def binaryRepr(self, datafile, value):
        return struct.pack(FloatParameter.structfmt, value)
    def binaryRead(self, parser):
        b = parser.getBytes(FloatParameter.structsize)
        (val,) = struct.unpack(FloatParameter.structfmt, b)
        return val
    def __repr__(self):
        return "FloatParameter(%s, value=%s, default=%s)" \
               % (`self.name`, `self.value`, `self.default`)
    def valueDesc(self):
        return "A real number."

class IntParameter(ParameterBase):
    def __init__(self, name, value=None, default=0, tip=None):
        ParameterBase.__init__(self, name, value=value, default=default,
                               tip=tip)
    def checker(self, x):
        if type(x) is not IntType:
            raiseTypeError(type(x), IntType)
    def binaryRepr(self, datafile, value):
        return struct.pack(structIntFmt, value)
    def binaryRead(self, parser):
        b = parser.getBytes(structIntSize)
        (val,) = struct.unpack(structIntFmt, b)
        return val
    def __repr__(self):
        return "IntParameter(%s,%s)" % (`self.name`, `self.value`)
    def valueDesc(self):
        return "Integer."

class ListOfIntsParameter(ParameterBase):
    def __init__(self, name, value=None, default=[], tip=None):
        ParameterBase.__init__(self, name, value, default, tip)
    def checker(self, x):
        if type(x) is not ListType:
            raiseTypeError(type(x), "list of Ints")
        for y in x:
            if type(y) is not IntType:
                raiseTypeError("list of %s" % type(y), "list of Ints")
    def binaryRepr(self, datafile, value):
        lengthstr = struct.pack(structIntFmt, len(value))
        datafmt = '>%di' % len(value)
        datastr = struct.pack(datafmt, *value)
        return lengthstr + datastr
    def binaryRead(self, parser):
        b = parser.getBytes(structIntSize)
        (length,) = struct.unpack(structIntFmt, b)
        format = '>%di' % length
        datasize = struct.calcsize(format)
        b = parser.getBytes(datasize)
        return list(struct.unpack(format, b))
    def valueDesc(self):
        return "A list of integers."

# The SkeletonInfo toolbox requires a tuple of ints parameter for
# which "None" is also a valid value.  
class TupleOfIntsParameter(ParameterBase):
    def __init__(self, name, value=None, default=None, tip=None):
        ParameterBase.__init__(self, name,
                               value=value, default=default, tip=tip)
    def checker(self, x):
        if x is not None:
            if type(x) is not TupleType:
                raiseTypeError(type(x), TupleType)
            for i in x:
                if type(i) is not IntType:
                    raisetypeError(
                        "Tuple of %s" % type(i), "Tuple of Ints.")
    def valueDesc(self):
        return "A tuple of integers, or None."
        

class ListOfUnsignedShortsParameter(ParameterBase):
    def __init__(self, name, value=None, default=[], tip=None):
        ParameterBase.__init__(self, name, value, default, tip)
    def checker(self, x):
        if type(x) is not ListType:
            raiseTypeError(type(x), "list of unsigned shorts")
        for y in x:
            if type(y) is not IntType:
                raiseTypeError("list of %s" % type(y),
                               "list of unsigned shorts")
    def binaryRepr(self, datafile, value):
        lengthstr = struct.pack(structIntFmt, len(value))
        datafmt = '>%dH' % len(value)
        datastr = struct.pack(datafmt, *value)
        return lengthstr + datastr
    def binaryRead(self, parser):
        b = parser.getBytes(structIntSize)
        (length,) = struct.unpack(structIntFmt, b)
        format = '>%dH' % length
        datasize = struct.calcsize(format)
        b = parser.getBytes(datasize)
        return list(struct.unpack(format, b))
    def valueDesc(self):
        return "A list of unsigned short integers."

class ListOfTuplesOfIntsParameter(ParameterBase):
    def __init__(self, name, value=None, default=[], tip=None):
        ParameterBase.__init__(self, name, value, default, tip)
    def checker(self, x):
        if type(x) is not ListType:
            raise TypeError("Expected a List of Tuples of Ints!")
        # Type-check doesn't take much longer than assignment.
        for t in x:
            if type(t) is not TupleType:
                raise TypeError("Expected a List of Tuples of Ints!")
            for o in t:
                if type(o) is not IntType:
                    raise TypeError("Expected a List of Tuples of Ints!")
    def binaryRepr(self, datafile, value):
        length = struct.pack(structIntFmt, len(value))
        strings = [length]
        for tpl in value:
            tlen = len(tpl)
            lengthstr = struct.pack(structIntFmt, tlen)
            data = struct.pack('>%di'% tlen, *tpl)
            strings.append(lengthstr)
            strings.append(data)
        return string.join(strings, '')
    def binaryRead(self, parser):
        b = parser.getBytes(structIntSize)
        (length,) = struct.unpack(structIntFmt, b)
        val = [None]*length
        for i in range(length):
            b = parser.getBytes(structIntSize)
            (tlen,) = struct.unpack(structIntFmt, b)
            format = '>%di' % tlen
            b = parser.getBytes(struct.calcsize(format))
            val[i] = struct.unpack(format, b)
        return val
    def valueRepr(self):
        return "List of variably sized Tuples of Ints"
    def valueDesc(self):
        return "A list of tuples of integers.  The tuples do not all have to have the same size."

class ListOfListOfIntsParameter(ParameterBase):
    # The binary repr for this parameter assumes that all the sublists
    # are the same length.
    def __init__(self, name, value=None, default=[], tip=None):
        ParameterBase.__init__(self, name, value, default, tip)
    def checker(self, x):
        if type(x) is not ListType:
            raise TypeError("Expected a List of Lists of Ints")
        for t in x:
            if type(t) is not ListType or len(t) != len(x[0]):
                raise TypeError("Expected a List of uniform Lists of Ints")
            for o in t:
                if type(o) is not IntType:
                    raise TypeError("Expected a List of Lists of Ints")
    def binaryRepr(self, datafile, value):
        h = len(value)
        w = len(value[0])
        size = struct.pack('>ii', h, w)
        strings = [size]
        for lst in value:
            strings.append(struct.pack('>%di' % w, *lst))
        return string.join(strings, '')
    def binaryRead(self, parser):
        b = parser.getBytes(struct.calcsize('>ii'))
        (h, w) = struct.unpack('>ii', b)
        format = '>%di' % w
        chunksize = struct.calcsize(format)
        val = [None]*h
        for i in range(h):
            b = parser.getBytes(chunksize)
            val[i] = list(struct.unpack(format, b))
        return val
    def valueRepr(self):
        return "List of uniformly sized Lists of Ints"
    def valueDesc(self):
        return "A list of lists of integers.  The sublists must all have the same size."
                
class ListOfTuplesOfFloatsParameter(ParameterBase):
    def __init__(self, name, value=None, default=[], tip=None):
        ParameterBase.__init__(self, name, value, default, tip)
    def checker(self, x):
        if type(x) is not ListType:
            raise TypeError("Expected a List of Tuples of Floats!")
        # Type-check doesn't take much longer than assignment.
        for t in x:
            if type(t) is not TupleType:
                raise TypeError("Expected a List of Tuples of Floats!")
            for o in t:
                if type(o) not in (FloatType, IntType):
                    raise TypeError("Expected a List of Tuples of Floats!")
    def binaryRepr(self, datafile, value):
        length = struct.pack(structIntFmt, len(value))
        strings = [length]
        for tpl in value:
            tlen = len(tpl)
            lengthstr = struct.pack(structIntFmt, tlen)
            data = struct.pack('>%dd' % tlen, *tpl)
            strings.append(lengthstr)
            strings.append(data)
        return string.join(strings, '')
    def binaryRead(self, parser):
        b = parser.getBytes(structIntSize)
        (length,) = struct.unpack(structIntFmt, b)
        val = [None]*length
        for i in range(length):
            b = parser.getBytes(structIntSize)
            (tlen,) = struct.unpack(structIntFmt, b)
            format = '>%dd' % tlen
            b = parser.getBytes(struct.calcsize(format))
            val[i] = struct.unpack(format, b)
        return val
    def valueRepr(self):
        return "List of variably sized Tuples of Floats"
    def valueDesc(self):
        return "A list of tuples of real numbers. The tuples do not all have to have the same size."

class ListOfTuplesOfIntFloatsParameter(ParameterBase):
    def __init__(self, name, value=None, default=[], tip=None):
        ParameterBase.__init__(self, name, value, tip)
    def checker(self, x):
        if type(x) is not ListType:
            raise TypeError("Expected a List of Tuples!")
        # Type-check doesn't take much longer than assignment.
        for t in x:
            if type(t) is not TupleType:
                raise TypeError("Expected Tuples in the List!")
            if type(t[0]) is not IntType:
                raise TypeError("Expected Int for the first slot in Tuple!")
            for o in t[1:]:
                if type(o) not in (FloatType, IntType):
                    raise TypeError("Expected Floats!")
    def binaryRepr(self, datafile, value):
        listlength = struct.pack(structIntFmt, len(value))
        strings = [listlength]
        for tpl in value:
            tlen = len(tpl)-1           # number of doubles in the tuple
            lengthstr = struct.pack(structIntFmt, tlen)
            data = struct.pack('>i%dd' % tlen, tpl[0], *(tpl[1:]))
            strings.append(lengthstr)
            strings.append(data)
        return string.join(strings, '')
    def binaryRead(self, parser):
        b = parser.getBytes(structIntSize)
        (listlength,) = struct.unpack(structIntFmt, b)
        val = [None]*listlength
        for i in range(listlength):
            b = parser.getBytes(structIntSize)
            (tlen,) = struct.unpack(structIntFmt, b)
            format = '>i%dd' % tlen
            b = parser.getBytes(struct.calcsize(format))
            val[i] = struct.unpack(format, b)
        return val
    def valueRepr(self):
        return "List of Tuples containing one Int followed by a variable number of Floats"
    def valueDesc(self):
        return "A list of tuples containing one integer followed by a variable number of real numbers."

# RegisteredParameter stores an instance of a RegisteredClass.
class RegisteredParameter(ParameterBase):
    def __init__(self, name, reg, value=None, default=None, tip=None):
        # reg must be a RegisteredClass class.
        self.registry = reg.registry
        self.reg = reg
        ParameterBase.__init__(self, name, value, default, tip)
    def checker(self, x):
        if type(x)==InstanceType:
            klass = x.__class__
        else:
            klass = x
        for registration in self.registry:
            if issubclass(klass, registration.subclass):
                return
        raise TypeError(
            'Bad type for RegisteredParameter! Got %s\nExpected one of %s'
            % (`klass`, `[reg.subclass for reg in self.registry]`))
    def clone(self):
        return self.__class__(self.name, self.reg, self.value, self.default,
                              self.tip)
    def regname(self):
        if type(self.reg)==ClassType:
            return self.reg.__name__
        # InstanceType.
        return self.reg.__class__
    def __repr__(self):
        return '%s(%s, %s, %s, %s)' %\
               (self.__class__.__name__, self.name, self.regname(),
                `self.value`, self.tip)

    # For registered class objects, the object itself has enough
    # information to write itself in binary, so the Parameter doesn't
    # have to do the work.  But for reading, there is no instance of
    # the RegisteredClass, so the parameter has to do it.
    def binaryRepr(self, datafile, value):
        return value.binaryRepr(datafile)
    def binaryRead(self, parser):
        return registeredclass.binaryReadRegClass(parser, self.registry)
    def classRepr(self):
        return "%s(%s)" % (self.__class__.__name__, self.regname())
    def valueRepr(self):
        return string.join([reg.subclass.__name__ for reg in self.registry],
                           '\n')
    def valueDesc(self):
        return \
         "An object of the <link linkend='RegisteredClass:%s'><classname>%s</classname></link> class." % \
               (self.regname(), self.regname())

# RegisteredListParameter stores a list of instances of
# RegisteredClasses.  No more than one member of each subclass can be
# present, although this restriction isn't enforced here.  The
# associated widget in GUI/regclassfactory.py does enforce it.

class RegisteredListParameter(RegisteredParameter):
    def __init__(self, name, reg, value=None, default=[], tip=None):
        # "value"s default value is different in RegisteredParameter.
        RegisteredParameter.__init__(self, name, reg, value, default, tip)
    def checker(self, x):
        if not (type(x)==ListType or type(x)==TupleType):
            raise TypeError('Type of RegisteredListParameter must be a List')
        for p in x:
            RegisteredParameter.checker(self, p)
    def binaryRepr(self, datafile, value):
        lengthstr = struct.pack(structIntFmt, len(value))
        strings = [lengthstr]
        for obj in value:
            strings.append(obj.binaryRepr(datafile))
        return string.join(strings, '')
    def binaryRead(self, parser):
        b = parser.getBytes(structIntSize)
        (length,) = struct.unpack(structIntFmt, b)
        val = [None]*length
        for i in range(length):
            val[i] = registeredclass.binaryReadRegClass(parser, self.registry)
        return val
    def valueDesc(self):
        return \
        "A list of objects of the <link linkend='RegisteredClass:%s'><classname>%s</classname></link> class." \
        % (self.regname(), self.regname())

# The AutomaticNameParameter can be set to either a string or the
# object automatic.automatic.  When asked for its value, however, it
# always returns a string, which is generated from the raw input by
# calling the provided 'resolver' function.  The arguments to the
# resolver are the parameter and the raw input.  The resolver can use
# the boolean function AutomaticNameParameter.automatic() to see if
# the raw input was automatic.automatic.

# In order for the resolver function to be useful,
# AutomaticNameParameters should probably be in ParameterGroups.  That
# gives the resolver access to other parameters in the group, which it
# can use to determine what names are available.

# Because the AutomaticNameParameter needs to intercept both the
# setting and getting of its value, it doesn't call
# ParameterBase.__init__.  It stores the raw input in
# self.truevalue.

class AutomaticNameParameter(ParameterBase):
    def __init__(self, name, resolver, value=None, default=None, tip=None):
        self.checker = TypeChecker(StringType, automatic.Automatic)
        self.name = name                # name of the parameter
        self.tip = tip
        self.default = default
        self.timestamp = timestamp.TimeStamp()
        self.group = None
        self.resolver = resolver
        if value is not None:
            self.set(value)
        else:
            self.truevalue=None

    def set(self, value):
        if value is not None:
            try:
                self.checker(value)
            except TypeError, msg:
                raise ParameterMismatch(
                    str(msg) + "\nfor AutomaticNameParameter: " + self.name)
        self.truevalue=value
        self.timestamp.increment()

    # Parent class __setattr__ is OK, it just calls set.
    
    def clone(self):
        return AutomaticNameParameter(self.name,
                                      self.resolver,
                                      self.truevalue,
                                      self.default,
                                      self.tip)
    def nontrivial(self):
        # The "trivial" values are either self.truevalue = None or "".
        # These are both logically false.
        return self.truevalue
        
    def automatic(self):
        return self.truevalue==automatic.automatic

    def __getattr__(self, attr):
        if attr=='value':
            return self.resolver(self, self.truevalue)
        raise AttributeError

    def __repr__(self):
        return 'AutomaticNameParameter(%s, resolver=%s, truevalue=%s, %s)' % \
               (self.name, self.resolver, self.truevalue, self.tip)
    def valueDesc(self):
        return \
          "A character string, or the variable <constant>automatic</constant>."

#############################

# Backwards compliant factory function for creating Parameters of the
# right type.  Modified to only work if the resulting parameter
# is a RegisteredParameter -- will not create generic parameters
# for particular types, but will instead throw an exception.

def Parameter(name, types, *args, **kwargs):
    klass = None
    if type(types) is InstanceType:
        klass = types.__class__
    elif type(types) is ClassType:
        klass = types
    else:
        raise ErrPyProgrammingError(
            "Unable to create parameter for type %s." % `types`)
    
    # ConvertibleRegisteredClass is a subclass of RegisteredClass,
    # so check that hierarchy first.
    if klass and issubclass(klass, registeredclass.ConvertibleRegisteredClass):
        return apply(ConvertibleRegisteredParameter,
                     (name, types)+args, kwargs)
    # Then look for "ordinary" registered classes or CRegistered classes.
    if klass and (issubclass(klass, registeredclass.RegisteredClass)
                  or (hasattr(klass, "registry")
                      and type(klass.registry)==ListType)):
        return apply(RegisteredParameter, (name, types)+args, kwargs)
        
    # Failure to find a parameter is an error.
    raise ErrPyProgrammingError(
        "Unable to find parameter for type %s." % `types`)
    
    
##############################
        
# Specialized parameter classes with specialized 'set' methods (and
# widget constructors in GUI mode).  Trivial classes here are extended
# in common.IO.GUI.parameterwidgets.

class ConvertibleRegisteredParameter(RegisteredParameter):
    pass

        
###############################################

# The ParameterGroup object.  This object presents a list-like
# interface for retrieval of parameters.  It also sets the "group"
# attribute of its member parameters, so they can look up parameters
# upon which they depend at __getitem__('value') time, if their value
# is context-sensitive.

class ParameterGroup:
    def __init__(self, *params):
        self.params = []
        for p in params:
            p.set_group(self)
            self.params.append(p)

    def __repr__(self):
        return "ParameterGroup(%s)" % `[p.name for p in self.params]`

    # Provide the list-like functionality.

    def append(self, param):
        param.set_group(self)
        self.params.append(param)

    def remove(self, param):
        i = self.params.index(param) # May raise ValueError.
        self.params[i].set_group(None)
        del self.params[i]

    def index(self, param):
        return self.params.index(param)

    # Operations in combination with lists always result in plain
    # ordinary lists.
    def __radd__(self, other):
        if type(other)==ListType:
            return other + self.params[:]
        if isinstance(other, ParameterGroup):
            return other.params[:] + self.params[:]
        raise TypeError("ParameterGroups can only be combined with ParameterGroups or lists.")

    def __add__(self, other):
        if type(other)==ListType:
            return self.params[:] + other
        if isinstance(other, ParameterGroup):
            return self.params[:] + other.params[:]
        raise TypeError("ParameterGroups can only be combined with ParameterGroups or lists.")

    def insert(self, pos, param):
        param.set_group(self)
        self.params.insert(pos, param)

    # Works by integer index or string key.
    def __getitem__(self, index):
        if type(index)==IntType:
            return self.params[index]
        elif type(index)==StringType:
            for p in self.params:
                if p.name==index:
                    return p
            else:
                raise IndexError("Key %s not found in ParameterGroup." % index)
        else:
            raise TypeError("Invalid type for ParameterGroup query.")

    def __setitem__(self, index, param):
        param.set_group(self)
        self.params[index] = param

    def __delitem__(self, index):
        self.params[index].set_group(self)
        del sef.params[index]

    def __len__(self):
        return len(self.params)

    def __del__(self):
        for p in self.params:
            p.set_group(None)
            


############################################

# Some parameter values may want "deep"comparison operations on them.
# RegisteredClass already does this, but not all values are
# registered.  May have unexpected side-effects if the instance
# attributes have non-simple values.

class Comparable:
    def __eq__(self,other):
        if other is None:
            return 0
        if self.__class__!=other.__class__:
            return 0
        for (k,v) in self.__dict__.items():
            if v!=other.__dict__[k]:
                return 0
        return 1
    def __ne__(self,other):
        return 1-self.__eq__(other)

#########################

# Parameters whose values are global objects that are stored in the
# OOF namespace (such as Fields & Equations) can be subclasses of
# ObjParameter.  ObjParameter only defines binaryRead and binaryRepr--
# subclasses must also be derived from ParameterBase.

class ObjParameter:
    def binaryRepr(self, datafile, value):
        classkey = datafile.oofObjID(value)
        return struct.pack('>i', classkey)
    def binaryRead(self, parser):
        (classkey,) = struct.unpack('>i',
                                    parser.getBytes(struct.calcsize('>i')))
        return parser.getObject(classkey)

#############################

# Parameters that don't need a tooltip should set their "tip" argument
# to emptyTipString.  This will prevent the api dump from complaining
# about a missing tip.

class EmptyTip:
    def __nonzero__(self):
        return 0
    def read(self, *args):              # for xmlmenudump.getHelp()
        return ""

emptyTipString = EmptyTip()
