# -*- python -*-
# $RCSfile: pixelgroupmenu.py,v $
# $Revision: 1.41 $
# $Author: langer $
# $Date: 2004/12/13 19:46:46 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# Menu commands for manipulating PixelGroups

from SWIG.common import ooferror
from SWIG.common import pixelgroup
from SWIG.common import switchboard
from common import debug
from common import primitives
from common.IO import automatic
from common.IO import microstructureIO
from common.IO import parameter
from common.IO import reporter
from common.IO import whoville
from common.IO import xmlmenudump
from common.IO.mainmenu import OOF
from common.IO.oofmenu import OOFMenuItem
from common.IO.pixelgroupparam import PixelGroupParameter
from types import *
import common.microstructure      # a local variable is named 'microstructure'

Parameter = parameter.Parameter
BooleanParameter = parameter.BooleanParameter
AutomaticNameParameter = parameter.AutomaticNameParameter
StringParameter = parameter.StringParameter

pixgrpmenu = OOF.addItem(OOFMenuItem('PixelGroup', cli_only=1))

##########################

def newPixelGroup(menuitem, name, microstructure):
    if name and microstructure:
        mscontext = common.microstructure.microStructures[microstructure]
        ms = mscontext.getObject()
        mscontext.begin_writing()
        try:
            if ms:
                grp = ms.getGroup(name)  # creates new group and issues
                                         # "new pixel group" signal
                return grp
        finally:
            mscontext.end_writing()
    reporter.report("Failed to create group", name, "in microstructure",
                    microstructure)

def pixelGroupNameResolver(param, startname):
    if param.automatic():
        basename = 'pixelgroup'
    else:
        basename = startname
    msname = param.group['microstructure'].value
    ms = common.microstructure.getMicrostructure(msname)
    return ms.uniqueGroupName(basename)

pixgrpmenu.addItem(OOFMenuItem(
    'New', callback=newPixelGroup,
    params=parameter.ParameterGroup(
    AutomaticNameParameter('name', value=automatic.automatic,
                           resolver=pixelGroupNameResolver,
                           tip="Group name."),
    whoville.WhoParameter('microstructure', whoville.getClass('Microstructure'),
                          tip=
                          "Microstructure in which to create this PixelGroup.")
    ),
    help='Create a new PixelGroup in the given Microstructure.',
    discussion="<para><link linkend='Section:Concepts:Microstructure:Pixel_Group'>Pixel Group</link> is one of key attributes of &micro;. It doesn't necessarily have to be used this way, but most of times, it is used to contain a bunch of pixels that represent a &material;.</para>"))

##########################

def renamePixelGroup(menuitem, microstructure, group, new_name):
    # "group" arg is the old group name.
    mscontext = common.microstructure.microStructures[microstructure]
    ms = mscontext.getObject()
    mscontext.begin_writing()
    try:
        grp = ms.findGroup(group)
        # Don't just say "if grp" here.  PixelGroup has a __len__
        # function, so empty groups evaluate to "false".
        if grp is not None:
            ms.renameGroup(group, new_name)
            switchboard.notify('renamed pixel group', grp, group, new_name)
        else:
            raise ooferror.ErrUserError("There is no pixel group named %s!"
                                        % group)
    finally:
        mscontext.end_writing()

pixgrpmenu.addItem(OOFMenuItem(
    'Rename', callback=renamePixelGroup,
    params=[
    whoville.WhoParameter('microstructure',
                          common.microstructure.microStructures,
                          tip=parameter.emptyTipString),
    PixelGroupParameter('group', tip='PixelGroup to be renamed.'),
    StringParameter('new_name', 
              tip='New name for the group, in quotation marks.')
    ],
    help='Rename an existing PixelGroup in the given Microstructure.',
    discussion="<para><emphasis>Did you know</emphasis>: If you unknowingly rename a <link linkend='Section:Concepts:Microstructure:Pixel_Group'>pixel group</link> with an exisiting name, &oof2; smartly catches it and make it unique.</para>"))

##########################

def copyPixelGroup(menuitem, microstructure, group, name):
    if group != name:
        mscontext = common.microstructure.microStructures[microstructure]
        ms = mscontext.getObject()
        mscontext.begin_writing()
        try:
            oldgroup = ms.findGroup(group)
            if oldgroup is not None:
                newgroup = ms.getGroup(name)
                newgroup.addWithoutCheck(oldgroup.members())
            else:
                raise ooferror.ErrUserError("There is no pixel group named %s!"
                                            % group)
        finally:
            mscontext.end_writing()
            
pixgrpmenu.addItem(OOFMenuItem(
    'Copy', callback=copyPixelGroup,
    params=parameter.ParameterGroup(
    whoville.WhoParameter('microstructure',
                          common.microstructure.microStructures,
                          tip=parameter.emptyTipString),
    PixelGroupParameter('group', tip='PixelGroup to be copied.'),
    AutomaticNameParameter('name', value=automatic.automatic,
                           resolver=pixelGroupNameResolver,
                           tip="Group name.")
    ),
    help='Make a copy of an existing pixel group',
    discussion="<para>Copy an exisiting <link linkend='Section:Concepts:Microstructure:Pixel_Group'>pixel group</link>.</para>"))

##########################

def destroyPixelGroup(menuitem, microstructure, group):
    mscontext = common.microstructure.microStructures[microstructure]
    ms = mscontext.getObject()
    mscontext.begin_writing()
    try:
        ms.removeGroup(group)  # "destroy pixel gorup" signal
    finally:
        mscontext.end_writing()

pixgrpmenu.addItem(OOFMenuItem(
    'Delete',
    callback=destroyPixelGroup,
    params=[
    whoville.WhoParameter('microstructure',
                          common.microstructure.microStructures,
                          tip=parameter.emptyTipString),
    PixelGroupParameter('group', tip='PixelGroup to be destroyed.')
    ],
    help='Delete the selected Pixel Group.',
    discussion="<para>Remove a <link linkend='Section:Concepts:Microstructure:Pixel_Group'>pixel group</link> completely from the system.</para>"))

##########################

def meshablePixelGroup(menuitem, microstructure, group, meshable):
    mscontext = common.microstructure.microStructures[microstructure]
    ms = mscontext.getObject()
    mscontext.begin_writing()
    try:
        grp = ms.findGroup(group)
        if grp is not None:
            grp.set_meshable(meshable)
            ms.recategorize()
        else:
            raise ooferror.ErrUserError("There is no pixel group named %s!"
                                        % group)
    finally:
        mscontext.end_writing()
        
pixgrpmenu.addItem(OOFMenuItem(
    'Meshable',
    callback=meshablePixelGroup,
    params=[
    whoville.WhoParameter('microstructure',
                          common.microstructure.microStructures,
                          tip=parameter.emptyTipString),
    PixelGroupParameter('group', tip="Pixel group."),
    BooleanParameter('meshable', tip="1 (true) for meshable and 0 (false) for non-meshable.")],
    help="Should adaptive Skeletons follow the boundaries of the given group?",
    discussion="<para>By default, a <link linkend='Section:Concepts:Microstructure:Pixel_Group'>pixel group</link> is meshable, meaning boundaries formed by the pixel gorup is respected by adaptive meshing tools -- most of <link linkend='MenuItem:OOF.Skeleton.Modify'>modification</link> tools for &skel;.</para>"))
        

##########################

def addSelection(menuitem, microstructure, group):
    mscontext = common.microstructure.microStructures[microstructure]
    ms = mscontext.getObject()
    mscontext.begin_writing()
    try:
        grp = ms.findGroup(group)
        ms.pixelselection.begin_reading()
        sel = ms.pixelselection.getObject()
        # check for size compatibility
        if ms.size()!=sel.size() or ms.sizeInPixels()!=sel.sizeInPixels():
            reporter.report(
                'Selection geometry does not match Microstructure!')
        else:
            grp.add(sel.members())  # calls ms.recategorize(), which increments
                                    # the timestamp of ms
                                    # AND issues "changed pixel group" signal
    finally:
        ms.pixelselection.end_reading()
        mscontext.end_writing()

pixgrpmenu.addItem(OOFMenuItem(
    'AddSelection',
    callback=addSelection,
    params=[
    whoville.WhoParameter('microstructure',
                          common.microstructure.microStructures,
                          tip=parameter.emptyTipString),
    PixelGroupParameter('group',
                        tip="Group to which to add the selected pixels.")
            ],
    help='Add the currently selected pixels to the given PixelGroup.',
    discussion="<para>The pixels that are currently <link linkend='Section:Concepts:Microstructure:PixelSelection'>selected</link> will be added to the selected <link linkend='Section:Concepts:Microstructure:Pixel_Group'>pixel group</link>.</para>"))
    
#########################

def removeSelection(menuitem, microstructure, group):
    mscontext = common.microstructure.microStructures[microstructure]
    ms = mscontext.getObject()
    mscontext.begin_writing()
    try:
        grp = ms.findGroup(group)
        ms.pixelselection.begin_reading()
        sel = ms.pixelselection.getObject()
        if sel is not None:
            grp.remove(sel.members())  # calls ms.recategorize(), which
                                       # increments the timestamp of ms AND
                                       # issues "changed pixel group" signal.
    finally:
        ms.pixelselection.end_reading()
        mscontext.end_writing()

pixgrpmenu.addItem(OOFMenuItem(
    'RemoveSelection',
    callback=removeSelection,
    params=[
    whoville.WhoParameter('microstructure',
                          common.microstructure.microStructures,
                          tip=parameter.emptyTipString),
    PixelGroupParameter('group',
                        tip="Group from which to remove the selected pixels.")
            ],
    help='Remove the currently selected pixels from the given PixelGroup.',
    discussion="<para>The pixels that are currently <link linkend='Section:Concepts:Microstructure:PixelSelection'>selected</link> and belong to the selected pixel group will be removed from the <link linkend='Section:Concepts:Microstructure:Pixel_Group'>pixel group</link>.</para>"))

#########################

def clearGroup(menuitem, microstructure, group):
    mscontext = common.microstructure.microStructures[microstructure]
    ms = mscontext.getObject()
    mscontext.begin_writing()
    try:
        grp = ms.findGroup(group)
        grp.clear()  # issues "changed pixel group" signal.
    finally:
        mscontext.end_writing()

pixgrpmenu.addItem(OOFMenuItem(
    'Clear',
    callback=clearGroup,
    params=[
    whoville.WhoParameter('microstructure',
                          common.microstructure.microStructures,
                          tip=parameter.emptyTipString),
    PixelGroupParameter('group', tip='Group from which to remove all pixels.')
    ],
    help="Remove all pixels from the given PixelGroup.",
    discussion="<para>Empty the selected <link linkend='Section:Concepts:Microstructure:Pixel_Group'>pixel group</link>.</para>"))

def queryGroup(menuitem, microstructure, group):
    mscontext = common.microstructure.microStructures[microstructure]
    ms = mscontext.getObject()
    mscontext.begin_reading()
    try:
        grp = ms.findGroup(group)
        nop = len(grp)
        areaOfGroup = nop*ms.areaOfPixels()
    finally:
        mscontext.end_reading()
    reporter.report(">>> ", nop, " pixels, ", "area = ", areaOfGroup)

pixgrpmenu.addItem(OOFMenuItem(
    'Query',
    callback=queryGroup,
    params=[
    whoville.WhoParameter('microstructure',
                          common.microstructure.microStructures,
                          tip=parameter.emptyTipString),
    PixelGroupParameter('group', tip='Get information on this group.')
    ],
    help="Query the given PixelGroup.",
    discussion="<para>Get some information from the <link linkend='Section:Concepts:Microstructure:Pixel_Group'>pixel group</link>.</para>"))
    
        
############################

# Functions for reading and writing pixelgroups in a data file. 

def _writeData(self, dfile, microstructure, pixel):
    grpnames = pixelgroup.pixelGroupNames(microstructure, pixel)
    if grpnames:
        dfile.argument('groups', grpnames)
        return 1
    return 0

pixelgroup.PixelGroupAttributeRegistrationPtr.writeData = _writeData

def _readPixelGroups(menuitem, microstructure, category, groups):
    mscontext = common.microstructure.microStructures[microstructure]
    ms = mscontext.getObject()
    mscontext.begin_writing()
    try:
        pixels = microstructureIO.getCategoryPixels(microstructure, category)
        for groupname in groups:
            grp = ms.getGroup(groupname)
            grp.add(pixels)
    finally:
        mscontext.end_writing()

microstructureIO.categorymenu.addItem(OOFMenuItem(
    pixelgroup.attributeReg.name(),
    callback=_readPixelGroups,
    params=[
    whoville.WhoParameter('microstructure',
                          common.microstructure.microStructures,
                          tip=parameter.emptyTipString),
    parameter.IntParameter('category', tip="Pixel category."),
    parameter.ListOfStringsParameter('groups', tip="List of names of pixel groups.")
    ],
    help="Assign pixel groups to pixel categories. Used internally in data files.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/common/IO/menu-pgroupcategory.xml')
    ))
