# -*- python -*-
# $RCSfile: analyze.py,v $
# $Revision: 1.30 $
# $Author: reida $
# $Date: 2004/12/15 20:44:39 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# Objects for performing operations on data.  These include writing
# data to a file, taking averages and standard deviations, and so
# forth.  The data to which this should be done is specified by means
# of an output.

from common import enum
from common import primitives
from common import registeredclass
from common.IO import destinationparam
from common.IO import parameter
from common.IO import whoville
import engine.mesh
import math
import string


# Parent class for all "Data Operation" objects, i.e. things which do
# things to outputs, like writing them gridwise to files,
# evaluating them along cross-sections, and so forth.

# Data operation objects need to specify at registration time
# whether they are "direct" or not -- a "direct" operation is one
# for which the value of the output at a given sample is directly
# output to the user.  Statistical operations are not direct.

class DataOperation(registeredclass.RegisteredClass):
    registry = []


Separator = enum.EnumClass("space", "comma")

separator_strings = {
    Separator("space") : " ",
    Separator("comma") : ", " }


# The most basic, a direct-output class that just writes out the
# data at all the samples. 
class DirectOutput(DataOperation):
    def __init__(self, comment_character, separator):
        self.comment_character = comment_character
        self.separator = separator
    def __call__(self, output, domain, sampling, destination):
        header = sampling.make_samples(domain)
        output_header = header + ["Value"]
        sep_string = separator_strings[self.separator]
        print >> destination, self.comment_character + \
              string.join(output_header, sep_string)
        for (s,v) in sampling.evaluate(domain, output):
            tags = s.columnData(header)
            if len(tags)==1:
                print >> destination, \
                      string.join(tags[0]+[`v`], sep_string)
            else:
                for (tag, vstring) in zip(tags, [`val` for val in v]):
                    print >> destination, \
                          string.join(tag+[vstring], sep_string)
    

# "Direct Output" is special, in that the corresponding auto-generated
# menu item is used directly in the meshcstoolboxGUI code -- if the
# name is changed here, it should be changed there also.
registeredclass.Registration(
    'Direct Output', DataOperation, DirectOutput, 1,
    params=[parameter.StringParameter("comment_character",
                                      "#", default="#",
                                      tip="Set the string at the start of the header line."),
            enum.EnumParameter("separator", Separator,
                               "comma", default="comma",
                               tip="Set the character between data entries.")],
    help="Write the data values directly.", direct=1)




# Integrates the output over each sample, and normalizes by
# the power=0 version.
class AverageOutput(DataOperation):
    def __call__(self, output, domain, sampling, destination):
        sampling.fill_sample_list(domain)
        integralsum = 0.0
        areasum = 0.0
        for v in sampling.integrate(domain, output, power=0):
            areasum += v[1]
        for v in sampling.integrate(domain, output):
            integralsum += v[1]
            
        result = integralsum/areasum

        print >> destination, `result`

        
registeredclass.Registration(
    "Average", DataOperation, AverageOutput, 2,
    tip="Average the data over all the samples.",
    direct=0)



# Standard deviation, square root of the integral of ( (x-<x>)^2 ),
# which is of course (<x^2>-<x>^2)^(1/2)
class StdDevOutput(DataOperation):
    def __call__(self, output, domain, sampling, destination):
        sampling.fill_sample_list(domain)
        integralsum = 0.0
        areasum = 0.0
        squaresum = 0.0

        for v in sampling.integrate(domain, output, power=0):
            areasum += v[1]
        for v in sampling.integrate(domain, output, power=1):
            integralsum += v[1]
        for v in sampling.integrate(domain, output, power=2):
            squaresum += v[1]

        result = math.sqrt( (squaresum/areasum) - (integralsum/areasum)**2 )

        print >> destination, `result`
            

registeredclass.Registration(
    "Standard Deviation", DataOperation, StdDevOutput, 3,
    tip="Compute the standard deviation of the data over the samples.",
    direct=0)
