# -*- python -*-
# $RCSfile: displaymethods.py,v $
# $Revision: 1.77 $
# $Author: reida $
# $Date: 2004/11/19 21:33:41 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# Simple display methods for Skeletons and Meshes.

from SWIG.common import ooferror
from common import color
from common import debug
from common import primitives
from common import utils
from common.IO import colormap
from common.IO import display
from common.IO import ghostgfxwindow
from common.IO import output
from common.IO import parameter
from common.IO import xmlmenudump
from common import registeredclass
from engine import skeletoncontext
from engine.IO import outputClones
from engine.IO import outputDefs
import SWIG.engine.material

FloatRangeParameter = parameter.FloatRangeParameter
IntRangeParameter = parameter.IntRangeParameter
Parameter = parameter.Parameter

class SkeletonDisplayMethod(display.DisplayMethod):
    def __init__(self):
        display.DisplayMethod.__init__(self)
    def polygons(self, themesh):
        return [el.perimeter() for el in themesh.element_iterator()]
    
class MeshDisplayMethod(display.DisplayMethod):
    # A display method that displays data from a Mesh at positions
    # determined by the mesh.  self.where is a PositionOutput.
    def __init__(self):
        display.DisplayMethod.__init__(self)
    def incomputable(self, gfxwindow):
        mesh = self.who().resolve(gfxwindow)
        return display.DisplayMethod.incomputable(self, mesh) \
               or self.where.incomputable(mesh)
    def polygons(self, themesh):
        edges = [element.perimeter() for element in themesh.element_iterator()]
        flatedges = utils.flatten(edges)
        # corners tells where on each edge to evaluate self.where
        corners = [[0.0]]*len(flatedges)
        # evaluate position output for all edges at once
        pgons = self.where.evaluate(themesh, flatedges, corners)
        # give the corner list the same structure as the edge list: a
        # list of lists, where each sublist is the list of corners of
        # an element.
        pgons = utils.unflatten(edges, pgons)
        return pgons

###########################
    
# It would be nice to have a single DisplayMethod that could display
# element edges for both a Skeleton and a Mesh.  But the Mesh needs a
# PositionOutput and the Skeleton doesn't, so it's not possible.

class EdgeDisplay:
    def draw(self, gfxwindow, device):
        themesh = self.who().getObject(gfxwindow)
        polygons = self.polygons(themesh)
        device.comment("EdgeDisplay")
        device.set_lineColor(self.color)
        device.set_lineWidth(self.width)
        for polygonset in polygons:
            device.draw_polygon(primitives.Polygon(polygonset))

class MeshEdgeDisplay(EdgeDisplay, MeshDisplayMethod):
    # EdgeDisplay draws the edges of the Elements
    def __init__(self, where, width=0, color=color.black):
        if where is not None:
            self.where = where.clone()
        else:
            self.where = None           # ?
        self.width = width
        self.color = color
        MeshDisplayMethod.__init__(self)

class SkeletonEdgeDisplay(EdgeDisplay, SkeletonDisplayMethod):
    def __init__(self, width=0, color=color.black):
        self.width = width
        self.color = color
        SkeletonDisplayMethod.__init__(self)

#elementEdge =
registeredclass.Registration('Element Edges',
                             display.DisplayMethod,
                             MeshEdgeDisplay,
                             ordering=0.0,
                             layerordering=display.Linear,
                             params=[
    output.PositionOutputParameter('where', outputDefs.originalPosition,
                            tip="Plot at displaced or original position?"),
    color.ColorParameter('color', color.black,
                         tip="Color of the displayed edges."),
    IntRangeParameter('width', (0,10), 0, tip="Line thickness, in pixels.")],
                             whoclasses = ('Mesh',),
                             tip="Draw the edges of Mesh Elements.",
                             discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/IO/meshedgedisplay.xml')
                             )
    
registeredclass.Registration('Element Edges',
                             display.DisplayMethod,
                             SkeletonEdgeDisplay,
                             ordering=0.0,
                             layerordering=display.Linear,
                             params=[
    color.ColorParameter('color', color.black,
                         tip="Color of the displayed edges."),
    IntRangeParameter('width', (0,10), 0, tip="Line thickness, in pixels.")],
                             whoclasses = ('Skeleton',),
                             tip="Draw the edges of Skeleton Elements.",
                             discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/IO/skeletonedgedisplay.xml')
                             )

######################

class PerimeterDisplay(MeshDisplayMethod):
    def __init__(self, where, width=0, color=color.black):
        self.where = where.clone()
        self.width = width
        self.color = color
        MeshDisplayMethod.__init__(self)
    def draw(self, gfxwindow, device):
        mesh = self.who().resolve(gfxwindow)
        femesh = mesh.getObject()
        device.comment("PerimeterDisplay")
        device.set_lineColor(self.color)
        device.set_lineWidth(self.width)
        for element in femesh.element_iterator():
            el_edges = element.perimeter()
            for edge in el_edges:
                if element.exterior(edge.startpt(), edge.endpt()):
                    pts = self.where.evaluate(femesh, [edge], [[0.0, 1.0]])
                    device.draw_segment(primitives.Segment(pts[0], pts[1]))

registeredclass.Registration('Perimeter',
                             display.DisplayMethod,
                             PerimeterDisplay,
                             ordering=1.0,
                             layerordering=display.SemiLinear,
                             params=[
    output.PositionOutputParameter('where', outputDefs.originalPosition,
                            tip="Plot at displaced or original position?"),
    color.ColorParameter('color', color.black,
                         tip=parameter.emptyTipString),
    IntRangeParameter('width', (0,10), 0,
                      tip="Line width.")
    ],
                             whoclasses = ('Mesh',),
                             tip="Outline the perimeter of the Mesh",
                             discussion=xmlmenudump.loadFile(
    'DISCUSSIONS/engine/IO/perimeterdisplay.xml')
             )

######################

class CenterFillDisplay:
    # CenterFillDisplay fills each Element with a solid color,
    # determined by the 'what' at the center of the Element.

    def incomputable(self, gfxwindow):
        mesh = self.who().resolve(gfxwindow)
        return self.what.incomputable(mesh) or \
               MeshDisplayMethod.incomputable(self, gfxwindow)
    def draw(self, gfxwindow, device):
        themesh = self.who().resolve(gfxwindow).mesh()
        device.comment("SolidFill")
        device.set_colormap(self.colormap)
        polygons = self.polygons(themesh)
        elements = tuple(themesh.element_iterator())
        evaluationpoints = [[el.center()] for el in elements]
        values = self.what.evaluate(themesh, elements, evaluationpoints)
        max_value = max(values)
        min_value = min(values)
        # HACK: If the values are equal, put the actual value in
        # the middle of an arbitrary range.
        if max_value==min_value:
            max_value += 1.0
            min_value -= 1.0
        for polygon, value in zip(polygons, values):
            if value is not None:
                device.set_fillColor((value-min_value)/(max_value-min_value))
                device.fill_polygon(primitives.Polygon(polygon))
        
class MeshCenterFillDisplay(CenterFillDisplay, MeshDisplayMethod): 
    def __init__(self, where, what, colormap=colormap.ThermalMap()):
        self.where = where.clone()
        self.what = what.clone()
        self.colormap = colormap
        MeshDisplayMethod.__init__(self)


registeredclass.Registration('Solid Fill',
                             display.DisplayMethod,
                             MeshCenterFillDisplay,
                             ordering=2.0,
                             layerordering=display.Planar,
                             params=[
    output.ScalarOutputParameter('what', tip='Quantity to be plotted'),
    output.PositionOutputParameter('where', outputDefs.originalPosition,
                            tip='Plot at original or displaced position?'),
    Parameter('colormap', colormap.ColorMap, colormap.ThermalMap(),
              tip="Fill color.")],
                             whoclasses = ('Mesh',),
                             tip="Quick and dirty contour plot.",
                             discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/IO/centerfilldisplay.xml')
                             )

# SkeletonCenterFillDisplay isn't useful unless there are Outputs that
# can be evaluated on a Skeleton.

##class SkeletonCenterFillDisplay(CenterFillDisplay, SkeletonDisplayMethod): 
##    def __init__(self, what, colormap=colormap.ThermalMap()):
##        self.what = what
##        self.colormap = colormap

##registeredclass.Registration('SolidFill',
##             display.DisplayMethod,
##             SkeletonCenterFillDisplay,
##             ordering=2.01,
##             params=[
##    ScalarOutputParameter('what', tip='Quantity to be plotted'),
##    Parameter('colormap', colormap.ColorMap, colormap.ThermalMap())],
##             whoclasses = ('Mesh')
##             )

######################

class MaterialDisplay:
    def draw(self, gfxwindow, device):
        device.comment("Material Color")
        themesh = self.who().resolve(gfxwindow).getObject()
        polygons = self.polygons(themesh)
        # colorcache is a dictionary of colors keyed by Material.  It
        # prevents us from having to call material.fetchProperty for
        # each element.
        colorcache = {}
        for polygon, material in zip(polygons, self.materials(themesh)):
            if material is not None:
                try:
                    # If material has been seen already, retrieve its color.
                    color = colorcache[material]
                except KeyError:
                    # This material hasn't been seen yet.
                    try:
                        colorprop = material.fetchProperty('Color')
                        color = colorprop.color()
                    except ooferror.ErrNoSuchProperty:
                        color = None
                    colorcache[material] = color
                if color is not None:
                    device.set_fillColor(color)
                    device.fill_polygon(primitives.Polygon(polygon))
    def getTimeStamp(self, gfxwindow):
        microstructure = self.who().resolve(gfxwindow).getMicrostructure()
        return max(display.DisplayMethod.getTimeStamp(self, gfxwindow),
                   SWIG.engine.material.getMaterialTimeStamp(microstructure))
                    
class SkeletonMaterialDisplay(MaterialDisplay, SkeletonDisplayMethod):
    def __init__(self):
        SkeletonDisplayMethod.__init__(self)
    def materials(self, themesh):
        return [element.material(themesh)
                for element in themesh.element_iterator()]

class MeshMaterialDisplay(MaterialDisplay, MeshDisplayMethod):
    def __init__(self, where):
        self.where = where.clone()
        MeshDisplayMethod.__init__(self)
    def materials(self, themesh):
        return [element.material()
                for element in themesh.element_iterator()]
    
    
registeredclass.Registration('Material Color',
                             display.DisplayMethod,
                             SkeletonMaterialDisplay,
                             layerordering=display.Planar,
                             ordering=0.1,
                             params=[],
                             whoclasses=('Skeleton',),
                             tip="Fill each Element with the color of its assigned Material.",
                             discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/IO/skeletonmaterialdisplay.xml')
                             )

registeredclass.Registration('Material Color',
                             display.DisplayMethod,
                             MeshMaterialDisplay,
                             ordering=0.11,
                             layerordering=display.Planar,
                             params=[
    output.PositionOutputParameter('where', outputClones.posOutput,
                           tip='Plot at original or displaced position?')],
                             whoclasses=('Mesh',),
                             tip="Fill each Element with the color of its assigned Material.",
                                                          discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/IO/meshmaterialdisplay.xml')
                             )

###########################################

ghostgfxwindow.DefaultLayer(skeletoncontext.skeletonContexts,
                            SkeletonEdgeDisplay())
