# -*- python -*-
# $RCSfile: materialmenu.py,v $
# $Revision: 1.43 $
# $Author: langer $
# $Date: 2004/12/13 19:46:46 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov.

# Menu for operations related to the MaterialsManager.  Provides
# for the selection and deselection of materials and properties,
# creation and deletion of new materials, and the population of
# materials with properties.  Parametrization of properties themselves
# is handled by the corresponding propertymenu, and saving and loading
# are handled by the "File" submenu of the main OOF menu. 

# These menu operations need not correspond to entries
# in a menu bar -- they're triggered from the GUI by operations
# on the non-menubar buttons of the MaterialsPane.

from SWIG.common import pixelgroup
from SWIG.common import switchboard
from common import debug
from common import enum
from common import pixelselection
from common import quit
from common import primitives
from common import utils
from common.IO import automatic
from common.IO import datafile
from common.IO import microstructureIO
from common.IO import oofmenu
from common.IO import parameter
from common.IO import pixelgroupparam
from common.IO import whoville
from common.IO import mainmenu
from common.IO import xmlmenudump
from engine import materialmanager
from engine.propertyregistration import AllProperties
import SWIG.engine.material
import common.IO.selectionplaceholder
import common.microstructure
import types

StringParameter = parameter.StringParameter
OOFMenuItem = oofmenu.OOFMenuItem
matmanager = materialmanager.materialmanager


###################

class MaterialParameter(parameter.StringParameter):
    def checker(self, x):
        if x not in materialmanager.getMaterialNames():
            raise TypeError("Expected a Material name!")

class ListOfMaterialsParameter(parameter.ListOfStringsParameter):
    def checker(self, x):
        if type(x) is not types.ListType:
            raise TypeError("Expected a list of Material names!")
        names = materialmanager.getMaterialNames()
        for n in x:
            if n not in names:
                raise TypeError("Expected a list of Material names!")

###################

_materialmenu = OOFMenuItem('Material',cli_only=1)

mainmenu.OOF.addItem(_materialmenu)


# Add a new material, with a name.

def _newmaterial(menuitem, name):
    matmanager.add(name)

def materialNameResolver(param, startname):
    if param.automatic():
        basename = 'material'
    else:
        basename = startname
    return utils.uniqueName(basename, materialmanager.getMaterialNames())

_materialmenu.addItem(OOFMenuItem('New',
                                  callback=_newmaterial,
                                  params=[
    parameter.AutomaticNameParameter('name', resolver=materialNameResolver,
                                     value=automatic.automatic,
                                     tip="Name of the material.")
    ],
                                  help="Create a new Material."))

# Delete the named material.
def _deletematerial(menuitem, name):
    matmanager.delete(name)

_materialmenu.addItem(OOFMenuItem('Delete',
                                  callback=_deletematerial,
                                  params=[
    parameter.StringParameter('name',
                              tip="Name of the material.")],
                                  help="Delete a Material."))


# Makes a copy of the current material, with a new name.
    
def _copymaterial(menuitem, name, new_name):
    matmanager.copy(name, new_name)

    
_materialmenu.addItem(OOFMenuItem('Copy',
                                  callback=_copymaterial,
                                  params=[
    parameter.StringParameter('name', tip="Name of the source material."),
    parameter.AutomaticNameParameter('new_name', resolver=materialNameResolver,
                                     value=automatic.automatic,
                                     tip="Name of the new material.")],
                                  help="Copy a Material."
                                  ))



# Property operations.  Properties can be added or removed.

# "property" is actually the name of the property.  It's done
# this way so it looks sensible to the user.
def _addprop(menuitem,name,property):
    matmanager.add_prop(name, property)
    switchboard.notify("material changed", name)
    # matmanager.sync_select()

_materialmenu.addItem(OOFMenuItem('Add_property',
                                  callback=_addprop,
                                  params=[
    parameter.StringParameter('name', tip="Name of the material."),
    parameter.StringParameter('property', tip="Name of the property.")],
                                  help='Add a Property to a Material.'))

# Remove the named property from the named material.
def _removeprop(menuitem, name, property):
    matmanager.remove_prop(name, property)
    switchboard.notify("material changed", name)

_materialmenu.addItem(OOFMenuItem('Remove_property',
                                  callback=_removeprop,
                                  params = [
    parameter.StringParameter('name', tip="Name of the material."),
    parameter.StringParameter('property', tip="Name of the property.")],
                                  help="Remove a Property from a Material."
                                  ))
                      
# Saving and loading are done via the "file" menu, so they're not here.

#####################

def _assignmat(menuitem, material, microstructure, pixels):
    themat = materialmanager.getMaterial(material)
    ms = common.microstructure.microStructures[microstructure].getObject()
    if pixels == common.IO.selectionplaceholder.allPlaceHolder:
        themat.assignToAllPixels(ms)
    else:
        if pixels == common.IO.selectionplaceholder.selectionPlaceHolder:
            group = ms.pixelselection.getSelectionAsGroup()
        else:
            group = ms.findGroup(pixels)
        if group:
            themat.assignToPixelGroup(ms, group)
    switchboard.notify("materials changed in microstructure", ms)
    switchboard.notify("redraw")

_materialmenu.addItem(OOFMenuItem('Assign',
                                  callback=_assignmat,
                                  params=[
    MaterialParameter('material', tip="Material to be assigned."),
    whoville.WhoParameter('microstructure',
                          common.microstructure.microStructures,
                          tip=parameter.emptyTipString),
    pixelgroupparam.PixelAggregateParameter('pixels', tip="Target pixels.")],
                                  help="Assign a Material to pixels in a Microstructure."
                                  ))

def _removemat(menuitem, microstructure, pixels):
    ms = common.microstructure.microStructures[microstructure].getObject()
    if pixels == common.IO.selectionplaceholder.allPlaceHolder:
        SWIG.engine.material.removeAllMaterials(ms)
    else:
        if pixels == common.IO.selectionplaceholder.selectionPlaceHolder:
            group = ms.pixelselection.getSelectionAsGroup()
        else:
            group = ms.findGroup(pixels)
        if group:
            SWIG.engine.material.removeMaterialFromPixels(ms, group)
    switchboard.notify("materials changed in microstructure", ms)
    switchboard.notify("redraw")


_materialmenu.addItem(OOFMenuItem('Remove',
                                  callback=_removemat,
                                  params=[
    whoville.WhoParameter('microstructure',
                          common.microstructure.microStructures,
                          tip=parameter.emptyTipString),
    pixelgroupparam.PixelAggregateParameter('pixels', tip="Target pixels.")],
                                  help="Remove Material assignments from pixels in a Microstructure."
                                  ))
#############################

# Read and write a Microstructure's MaterialAttributes in a data file.

def _writeData(self, dfile, microstructure, pixel):
    material = SWIG.engine.material.getMaterialFromPoint(microstructure, pixel)
    if material is not None:
        dfile.argument('material', material.name())
        return 1
    return 0
    
SWIG.engine.material.MaterialAttributeRegistrationPtr.writeData = _writeData

def _readMSMaterial(menuitem, microstructure, category, material):
    if material is not None:
        mscontext = common.microstructure.microStructures[microstructure]
        ms = mscontext.getObject()
        mscontext.begin_writing()
        try:
            mat = materialmanager.getMaterial(material)
            pxls = microstructureIO.getCategoryPixels(microstructure, category)
            mat.assignToPixels(ms, pxls)
        finally:
            mscontext.end_writing()

def _writeGlobalData(self, dfile, microstructure):
    # Write the material definitions to the data file, so that they're
    # available when the pixel attributes are written.
    writeMaterials(dfile, SWIG.engine.material.getMaterials(microstructure))

SWIG.engine.material.MaterialAttributeRegistrationPtr.writeGlobalData = \
                                                              _writeGlobalData
        
microstructureIO.categorymenu.addItem(OOFMenuItem(
    SWIG.engine.material.attributeReg.name(),
    callback=_readMSMaterial,
    params=[
    whoville.WhoParameter("microstructure",
                          common.microstructure.microStructures,
                          tip=parameter.emptyTipString),
    parameter.IntParameter('category', tip="Category of pixels."),
    parameter.StringParameter('material', tip="Name of the Material.")
    ],
    help="Assign a Material to a pixel category. Used internally in data files.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/IO/menu-materialcategory.xml')
    ))

###############################

# Read and write a Material and its Properties in a data file.

def writeMaterials(dfile, materials):
    # Make sure to save each property only once, even if it's in more
    # than one Material.
    props = {}           # dictionary of all properties used in the Materials
    for material in materials:
        for prop in material.properties():
            props[prop.registration().name()] = prop
    # Sort the properties by name.
    paths = props.keys()
    paths.sort()
    # Write the properties.
    for path in paths:
        props[path].registration().writeData(dfile)
    # Write the materials.
    for material in materials:
        dfile.startCmd(mainmenu.OOF.LoadData.Material)
        dfile.argument('name', material.name())
        dfile.argument('properties', [prop.registration().name()
                                      for prop in material.properties()])
        dfile.endCmd()
        
def saveMaterials(menuitem, filename, mode, format, materials):
    quit.disallow()
    try:
        dfile = datafile.writeDataFile(filename, mode, format)
        writeMaterials(dfile,
                       [materialmanager.getMaterial(m) for m in materials])
        dfile.close()
    finally:
        quit.allow()

mainmenu.OOF.File.Save.addItem(oofmenu.OOFMenuItem(
    'Materials',
    callback = saveMaterials,
    params=[
    parameter.StringParameter('filename', tip="Name of the file."),
    parameter.StringParameter('mode', tip="'w' to (over)write, 'a' to append."),
    enum.EnumParameter('format', datafile.DataFileFormat, datafile.ASCII,
                       tip="Format of the file."),
    ListOfMaterialsParameter('materials', tip="Material(s) to be saved.")
    ],
    help="Save Materials to a file.",
    discussion="Available file formats are listed in <xref linkend='Table:data_option'/>."))


def loadMaterial(menuitem, name, properties):
    props = [AllProperties[propname]()  # create instances
             for propname in properties]
    materialmanager.materialmanager.new_material(name, *props)


mainmenu.OOF.LoadData.addItem(oofmenu.OOFMenuItem(
    'Material',
    callback = loadMaterial,
    params=[
    parameter.StringParameter('name', tip="Name of the material."),
    parameter.ListOfStringsParameter('properties', tip="Properties for the material.")
    ],
    help="Load Materials from a data file.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/IO/menu-loadmaterial.xml')
    ))
