# -*- python -*-
# $RCSfile: outputDefs.py,v $
# $Revision: 1.56 $
# $Author: langer $
# $Date: 2004/11/18 15:44:23 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# This file generates concrete Outputs from the Outputs defined in
# outputClones.py.  The concrete Outputs are stored in a tree
# structure which can easily be converted to a menu.

from SWIG.common import switchboard
from SWIG.engine import field
from SWIG.engine import flux
from SWIG.engine import outputval
from SWIG.engine import planarity
from SWIG.engine.IO import propertyoutput
from common import debug
from common import enum
from common import primitives
from common.IO import output
from common.IO import parameter
from common.IO import xmlmenudump
from engine import problem
from engine.IO import outputClones

from types import *
import string

######################################

# Displacement

# Although Displacement is a Field, it's convenient to treat it as a
# Point so that it can be added to a position.  Defining
# DisplacementOutput in terms of another output which sets a parameter
# to problem.Displacement allows the computability checking to work.

## TODO: The displacement output should be computable even if the
## Displacement field isn't defined.  It should just evaluate to
## (0,0).

Displacement = field.getCompoundField("Displacement")
iter = Displacement.iterator(planarity.IN_PLANE)
disp0 = iter.cloneIndex()
iter.next()
disp1 = iter.cloneIndex()

def _disp2point(fieldlist, *args, **kwargs):
    # Convert displacement field OutputVals to Points
    return [primitives.Point(field[disp0], field[disp1]) for field in fieldlist]

displacementFieldOutput = outputClones.FieldAnywayOutput.clone(
    name="displacement field")
displacementFieldOutput.plugParameter(which="field", value=Displacement)
displacementOutput = output.Output(name="dispPoint",
                                   oType=primitives.Point,
                                   func=_disp2point,
                                   inputs=output.Input("displacement",
                                                       outputval.OutputValPtr,
                                                       displacementFieldOutput))

##def _list2point(list, *args, **kwargs):
##    return [primitives.Point(p[0], p[1]) for p in list]

##List2PointOutput = \
##                 output.Output(name="list2point", func=_list2point,
##                               oType=primitives.Point,
##                               inputs=(output.Input("vector",
##                                            parameter.IndexableTypeChecker()),))

##displacementFieldOutput = \
##                outputClones.VectorFieldOutput.clone(name="displacement field")
##displacementFieldOutput.plugParameter(which='field', value=Displacement)

##displacementOutput = List2PointOutput.clone(name="displacement")
##displacementOutput.plugInput(input=displacementFieldOutput, which="vector")


# Enhanced position is scalefactor*Displacement + original position
enhancedPosition = outputClones.PointSumOutput.clone(
    name='enhanced position',
    tip='Exaggerated displacement field.',
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/IO/enhancedPosOutput.xml'))
enhancedPosition.plugInput(displacementOutput, 'point1')
enhancedPosition.plugInput(outputClones.posOutput, 'point2')
enhancedPosition.plugParameter(value=1.0, which='b')
enhancedPosition.plugParameter(default=1.0, which='a', rename='factor',
                               tip='Displacement multiplier.')

# Actual position is enhanced position with a scalefactor of 1
actualPosition = enhancedPosition.clone(
    name='actual position',
    tip='Displaced position.',
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/IO/actualPosOutput.xml')
    )
actualPosition.plugParameter(value=1.0, which='factor')

originalPosition = outputClones.posOutput.clone(
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/IO/originalPosOutput.xml'))

output.definePositionOutput('original', outputClones.posOutput)
output.definePositionOutput('actual', actualPosition)
output.definePositionOutput('enhanced', enhancedPosition)

######################

output.defineAggregateOutput('Field:Value', outputClones.FieldOutput)
output.defineAggregateOutput('Field:Derivative', outputClones.FieldDerivOutput)
output.defineAggregateOutput('Field:Invariant',
                             outputClones.FieldInvariantOutput)
output.defineAggregateOutput('Flux:Value', outputClones.FluxOutput)
output.defineAggregateOutput('Flux:Invariant', outputClones.FluxInvariantOutput)

output.defineScalarOutput('Field:Component', outputClones.FieldCompOutput)
output.defineScalarOutput('Field:Invariant', outputClones.FieldInvariantOutput)
output.defineScalarOutput('Field:Derivative:Component',
                          outputClones.FieldDerivCompOutput)
output.defineScalarOutput('Field:Derivative:Invariant',
                          outputClones.FieldDerivInvariantOutput)
output.defineScalarOutput('Flux:Component', outputClones.FluxCompOutput)
output.defineScalarOutput('Flux:Invariant', outputClones.FluxInvariantOutput)

from common import strfunction
xyfunc = outputClones.FunctionOutput.clone(
    tip='Compute an arbitrary function of x and y.',
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/IO/funcOutput.xml'))
xyfunc.plugParameter(default=strfunction.XYStrFunction('1.0'), which='f')
output.defineScalarOutput('XYFunction', xyfunc)

######################

# PropertyOutputs, which are computed by the Property objects in a
# Material, only have to be identified by name, type, and (possibly)
# initializer here.  The computations are all done by the
# PropertyOutput<TYPE> classes and the various Property::output virtual
# functions.

class EnergyType(enum.EnumClass(
    ("Total", "All contributions to the energy"),
    ("Elastic", "Elastic energy-- one half stress times geometric strain.")
    )): pass

propertyoutput.ScalarPropertyOutputRegistration(
    "Energy",
    parameters=enum.EnumParameter("etype", EnergyType, default="Total",
                                  tip='The type of energy to compute.'),
    tip='Compute an energy density.',
    discussion="""
      <para>Compute the energy density of the &fields; on the &mesh;.
      Different values of <varname>etype</varname> include different
      contributions to the energy.</para>"""
    )

