// -*- C++ -*-
// $RCSfile: heatconductivity.C,v $
// $Revision: 1.26 $
// $Author: redwing $
// $Date: 2004/11/18 16:06:55 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

// heat conductivity

#include <oofconfig.h>
#include "common/coord.h"
#include "engine/element.h"
#include "engine/elementnodeiterator.h"
#include "engine/field.h"
#include "engine/flux.h"
#include "engine/property/heatconductivity/heatconductivity.h"
#include "engine/fieldindex.h"
#include "engine/material.h"
#include "engine/property/orientation/orientation.h"
#include "common/space.h"
#include "engine/nodalequation.h"
using namespace std;

HeatConductivityProp::HeatConductivityProp(PyObject *registry, 
				   const string &nm)
  : Property(nm,registry),
    conductivitytensor(3,3)
{
  temperature = dynamic_cast<ScalarField*>(Field::getField("Temperature"));
  heat_flux = dynamic_cast<VectorFlux*>(Flux::getFlux("Heat_Flux"));
}

void HeatConductivityProp::precompute() {
}

int HeatConductivityProp::K_integration_order(int degree) const {
  return 2*degree;
}

void HeatConductivityProp::fluxmatrix(const FEMesh *mesh, const Element *el,
				      const ElementFuncNodeIterator &j,
				      const Flux *flux, FluxData *fluxdata,
				      const MasterPosition &pt)
  const
{
  // The heat flux matrix M_{ij} multiplies the vector of nodal
  // temperatures to give the vector heat current J at point pt.
  // M_{ij} = kappa_{ik} grad_k N_j
  // J_i = M_{ij} T_j
  // where N_j is the shapefunction at node j.

  if(*flux != *heat_flux) {
    throw ErrProgrammingError("Unexpected flux", __FILE__, __LINE__);
  }

  double sf = j.shapefunction(pt);
  double dsf0 = j.dshapefunction(0, pt);
  double dsf1 = j.dshapefunction(1, pt);

  // Loop over flux components.  Loop over all components, even if
  // the flux is in-plane, because the out-of-plane components of
  // the flux matrix are used to construct the constraint equation.
  for(VectorFieldIterator i; !i.end(); ++i) {
    // in-plane temperature gradient contributions
    fluxdata->matrix_element(mesh, i, temperature, j) -=
      conductivitytensor(i.integer(), 0) * dsf0 +
      conductivitytensor(i.integer(), 1) * dsf1;
    
    // out-of-plane temperature gradient contribution
    if(!temperature->in_plane(mesh)) {
      fluxdata->matrix_element(mesh, i, temperature->out_of_plane(), j)
	-= conductivitytensor(i.integer(), 2) * sf;
    }
  }
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

IsoHeatConductivityProp::IsoHeatConductivityProp(PyObject *reg,
					 const string &nm, double kppa)
  : HeatConductivityProp(reg, nm),
    kappa_(kppa)
{
}  

void IsoHeatConductivityProp::precompute() {
  HeatConductivityProp::precompute();
  conductivitytensor.set(0.0);
  conductivitytensor(0,0) = conductivitytensor(1,1)
    = conductivitytensor(2,2) = kappa_;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

AnisoHeatConductivityProp::AnisoHeatConductivityProp(PyObject *reg,
						     PyObject *self,
						     const string &nm,
						     SymmMatrix3 *k)
  : kappa_(*k),
    HeatConductivityProp(reg,nm),
    PythonNative<Property>(self)
{
}

void AnisoHeatConductivityProp::cross_reference(Material *mat) {
  orientation = 
       dynamic_cast<OrientationProp*>(mat->fetchProperty("Orientation"));
}

void AnisoHeatConductivityProp::precompute() {
  HeatConductivityProp::precompute();
  Trace("AnisoHeatConductivityProp::precompute");
  MATRIX_D c(3,3);
  for(int i=0;i<3;i++)
    for(int j=0;j<3;j++)
      c(i,j)=kappa_(i,j);

  conductivitytensor = orientation->rotate(c);
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//
//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

HeatSourceProp::HeatSourceProp(PyObject *reg, 
			       const std::string &name, double qd)
  : Property(name,reg),
    qdot_(qd) {
    heat_flux = dynamic_cast<VectorFlux*>(Flux::getFlux("Heat_Flux"));
}

// Since the added heat is uniform over the element, integration order is
// just that of the shapefunctions themselves.
int HeatSourceProp::K_integration_order(int n) const {
  return n;
}

// HeatSource adds to the right-hand side of the heat equation, 
// so we want fluxrhs and not fluxmatrix.
void HeatSourceProp::fluxrhs(const FEMesh *mesh, const Element *element,
			     const Flux *flux, FluxData *fluxdata,
			     const MasterPosition &masterpos) const {

  if(*flux != *heat_flux) {
    throw ErrProgrammingError("Unexpected flux.", __FILE__, __LINE__);
  }

  // In fact, there is only one component, so this is excessive.
  for(int i=0;i<heat_flux->divergence_dim();i++) {
    fluxdata->rhs_element(i) += qdot_;
  }
}
