// -*- C++ -*-
// $RCSfile: thermalexpansion.C,v $
// $Revision: 1.31 $
// $Author: redwing $
// $Date: 2004/11/18 16:06:55 $


/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>

#include "common/coord.h"
#include "engine/IO/propertyoutput.h"
#include "engine/element.h"
#include "engine/femesh.h"
#include "engine/flux.h"
#include "engine/material.h"
#include "engine/matvec.h"
#include "engine/property/elasticity/cijkl.h"
#include "engine/property/elasticity/elasticity.h"
#include "engine/property/orientation/orientation.h"
#include "thermalexpansion.h"



ThermalExpansionProp::ThermalExpansionProp(PyObject *reg, 
					   PyObject *self,
					   const std::string &nm, 
					   SymmMatrix3 *alpha, double t0)
  : Property(nm,reg),
    alpha_(*alpha),
    PythonNative<Property>(self),
    tzero_(t0)
{
  temperature=dynamic_cast<ScalarField*>(Field::getField("Temperature"));
  stress_flux=dynamic_cast<SymmetricTensorFlux*>(Flux::getFlux("Stress"));
}

void ThermalExpansionProp::cross_reference(Material *mat) {
  // find out which property is the elasticity
  elasticity = dynamic_cast<Elasticity*>(mat->fetchProperty("Elasticity"));
  orientation = dynamic_cast<OrientationProp*>
    (mat->fetchProperty("Orientation"));
}

void ThermalExpansionProp::precompute() {
  lab_alpha_ = alpha_.transform(orientation->eulerangle());
};

int ThermalExpansionProp::K_integration_order(int degree) const {
  return 2*degree - 1;
}


void ThermalExpansionProp::fluxmatrix(const FEMesh *mesh,
				      const Element *element,
				      const ElementFuncNodeIterator &nu, 
				      const Flux *flux, FluxData *fluxdata,
				      const MasterPosition &x) const
{
  // The stress-free strain is: 
  //                    [ alpha11*T   0      0  ]
  //     u_{kl}^0(T) =  [    0    alpha22*T  0  ]
  //                    [    0    0   alpha33*T ]
  //
  // Our goal is to subtract C_{ijkl}*u_{kl}^0(T) from the stress
  // tensor.
  if(*flux != *stress_flux) {
    throw ErrProgrammingError("Unexpected flux." __FILE__, __LINE__);
  }
  
//   Trace("ThermalExpansionProp::fluxmatrix");

  // Assume linear elasticity, for now. 
  const Cijkl modulus = elasticity->cijkl(mesh, element, x);

  double sfval = nu.shapefunction(x);
  // Loop over all tensor components.
  for(SymTensorIterator ij; !ij.end(); ++ij) {
    double &mtx_el = fluxdata->matrix_element(mesh, ij, temperature, nu); //returns flux matrix
    // Strain tensor components.
    for(SymTensorIterator kl; !kl.end(); ++kl) {
      double ca = modulus(ij, kl)*lab_alpha_(kl.row(), kl.col())*sfval;
      if(kl.diagonal()) {
	mtx_el -= ca;
      }
      else {
	mtx_el -= 2.0*ca;
      }
    }
  }
}

void ThermalExpansionProp::fluxrhs(const FEMesh *mesh, const Element *element,
				   const Flux *flux, FluxData *fluxdata,
				   const MasterPosition &x) const {
  
  if(*flux!=*stress_flux) {
    throw ErrProgrammingError("Unexpected flux." __FILE__, __LINE__);
  }
  const Cijkl modulus = elasticity->cijkl(mesh, element, x);

  for(SymTensorIterator ij; !ij.end(); ++ij) {
    double &offset_el = fluxdata->offset_element(mesh, ij); // reference!
    for(SymTensorIterator kl; !kl.end(); ++kl) {
      if(kl.diagonal()) {
	offset_el -= modulus(ij,kl)*lab_alpha_(kl.row(),kl.col())*tzero_;
      }
      else {
	offset_el -= 2.0*modulus(ij,kl)*lab_alpha_(kl.row(),kl.col())*tzero_;
      }
    }
  }
}


void ThermalExpansionProp::output(const FEMesh *mesh,
				  const Element *element,
				  const PropertyOutput *output,
				  const MasterPosition &pos,
				  OutputVal *data)
  const 
{
  const std::string &outputname = output->name();
  if(outputname == "Strain") {
    // The parameter is a Python StrainType instance.  Extract its name.
    acquirePyLock();
    std::string stype;
    try {
      PyObject *str = PyObject_CallMethod(output->parameters()[0], "name", 0);
      stype = PyString_AsString(str);
      Py_XDECREF(str);
    }
    catch(...) {
      releasePyLock();
      throw;
    }
    releasePyLock();
    SymmMatrix3 *sdata = dynamic_cast<SymmMatrix3*>(data);
    // Compute alpha*T with T interpolated to position pos
    // TODO LATER: This is baroque and may be slow.  Make a
    // specialized function for getting the value of a scalar field.
    const OutputValue tfield = element->outputField(mesh, *temperature, pos);
    const ScalarOutputVal *tval = 
      dynamic_cast<const ScalarOutputVal*>(tfield.valuePtr());
    double t = tval->value();

    if(stype == "Thermal")
      *sdata += lab_alpha_*t;
    else if(stype == "Elastic")
      *sdata -= lab_alpha_*t;
  }
}
