# -*- python -*-
# $RCSfile: refine.py,v $
# $Revision: 1.83 $
# $Author: langer $
# $Date: 2004/11/12 22:16:31 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 


# Refinement works in two stages.  First a CheckAndMark object (from
# checkandmark.py) checks the elements and marks edges for
# subdivision.  Then a RefinementRuleSet (from refinemethod.py) is
# used to actually refine the elements, according to how many times
# each edge is marked.

from common import debug
from common import enum
from common import registeredclass
from common import utils
from common.IO import parameter
from engine import checkandmark
from engine import skeletonmodifier
from engine import refinemethod
from engine import skeletonsegment
from engine import refinequadbisection
from SWIG.common.IO import progressbar

SkeletonSegment = skeletonsegment.SkeletonSegment

arbitrary_factor = 128          # used by findSignature().  Must be
				# larger than the largest number of
				# refinements of an edge.

################################

# The RefinementDegree classes specify how many times edges are
# divided initially by the CheckAndMark object, and also provide a
# 'markExtras' routine that marks additional edges for division after
# the check and mark stage.

# Changes as of Mar 04.
# Refinement Degree:
# - TriSection
# - BiSection
# Each degree has two rule sets.
# - Conservative: preserves topology, that is, if you refine a quad, you'll
#                 get only quads.
# - Liberal: MAY break topology

class RefinementDegree(registeredclass.RegisteredClass):
    registry = []
    def markExtras(self, skeleton, markedEdges):
        pass

RefinementDegree.tip = "Number of subdivitions per segment."
RefinementDegree.discussion = "<para>TODO: Add discussion.</para>"

class Bisection(RefinementDegree):
    divisions = 1
    def __init__(self, rule_set):
        self.rule_set = rule_set
    def markExtras(self, skeleton, markedEdges):
        if self.rule_set.name == "conservative":
            refinequadbisection.markExtras(skeleton, markedEdges)
        
registeredclass.Registration('Bisection', RefinementDegree,
                             Bisection, 1,
                             params = [
    enum.EnumParameter('rule_set', refinemethod.RuleSet,
                       refinemethod.conservativeRuleSetEnum(),
                       tip='How to subdivide elements')
    ],
                             tip="Divide element edges into two.")

class Trisection(RefinementDegree):
    divisions = 2
    def __init__(self, rule_set):
        self.rule_set = rule_set

registeredclass.Registration('Trisection', RefinementDegree,
                             Trisection, 0,
                             params = [
    enum.EnumParameter('rule_set', refinemethod.RuleSet,
                       refinemethod.conservativeRuleSetEnum(),
                       tip='How to subdivide elements')
    ],
                             tip="Gallia est omnis divisa in partes tres, as are the edges of the elements.")

#################################

class EdgeMarkings:
    # Class for storing and returning how many divisions should be
    # performed on an edge.  Edges are defined by a pair of nodes.
    def __init__(self):
        self.markings = {}
    def mark(self, node0, node1, ndivs):
        # arguments are the nodes defining the edge, and the number of
        # new nodes to add to that edge.
        key = utils.canonicalorder(node0, node1)
        try:
            # Only mark an edge if it's not already marked for more
            # divisions.
            if self.markings[key] < ndivs:
                self.markings[key] = ndivs
        except KeyError:
            self.markings[key] = ndivs

    def getMark(self, node0, node1):
        key = utils.canonicalorder(node0, node1)
        try:
            return self.markings[key]
        except KeyError:
            return 0

    def getMarks(self, element):
        nnodes = element.nnodes()
        return [self.getMark(element.nodes[i], element.nodes[(i+1)%nnodes])
                for i in range(nnodes)]

    def getNMarkedEdges(self, element):
        marks = self.getMarks(element)
        nmarks = 0
        for m in marks:
            if m: nmarks += 1
        return nmarks

class RefinementCriterion(registeredclass.RegisteredClass):
    registry = []

RefinementCriterion.tip = "Criteria to determine whether to accept or reject the current refinement."
RefinementCriterion.discussion = "<para>TODO: Add discussion.</para>"


class Unconditionally(RefinementCriterion):
    def __call__(self, skeleton, element):
        return 1
registeredclass.Registration('Unconditional', RefinementCriterion,
                             Unconditionally,
                             ordering=0,
                             tip='Refine any element that meets the target condition.')

class AreaUnits(enum.EnumClass(
    ('Pixel', 'Area is in units of the pixel area.'),
    ('Physical', 'Area is in physical units'))):
    pass

AreaUnits.tip = "Units of area. See <link linkend='RegisteredClass:MinimumArea'><classname>MinimumArea</classname></link> for details."
AreaUnits.discussion = "<para>TODO: Add discussion.</para>"

class MinimumArea(RefinementCriterion):
    def __init__(self, threshold, units):
        self.threshold = threshold
        self.units = units
    def __call__(self, skeleton, element):
        if self.units == 'Pixel':
            return element.area() > self.threshold*skeleton.MS.areaOfPixels()
        else:
            return element.area() > self.threshold

registeredclass.Registration('Minimum Area', RefinementCriterion,
                             MinimumArea,
                             ordering=1,
                             params=[
    parameter.FloatParameter('threshold', 10,
                             tip="Minimum acceptable element area."),
    enum.EnumParameter('units', AreaUnits, AreaUnits('Pixel'),
                       tip='Units for the minimum area')],
                             tip='Only refine elements with area greater than the given threshold.')

###################################
            
class Refine(skeletonmodifier.SkeletonModifier):
    def __init__(self, targets, criterion, degree):
        self.targets = targets      # CheckAndMark instance
        self.criterion = criterion  # Criterion for refinement
        self.degree = degree        # RefinementDegree instance
        # Available rules for the specified rule_set of refinement degree
        self.rules = refinemethod.getRuleSet(self.degree.rule_set.name)
    #########
        
    def apply(self, skeleton, context):
        # pBar.set_message("marking elements...       ")
        pBar = progressbar.getProgress()
        # Copy the old skeleton, without copying the elements.
        newSkeleton = skeleton.improperCopy()
        markedEdges = EdgeMarkings()
        self.newEdgeNodes = {}          # allows sharing of new edge nodes

        # Primary marking
        self.targets(skeleton, context, self.degree.divisions, markedEdges,
                     self.criterion)
        # Additional marking
        self.degree.markExtras(skeleton, markedEdges)

        # Refine elements and segments
        segmentdict = {}                # which segments have been handled
        elements = skeleton.elements
        n = len(elements)
        
        for ii in range(n):
            oldElement = elements[ii]
            oldnnodes = oldElement.nnodes()
            # Get list of number of subdivisions on each edge ("marks")
            marks = markedEdges.getMarks(oldElement)
            # Find the canonical order for the marks. (The order is
            # ambiguous due to the arbitrary choice of the starting
            # edge.  Finding the canonical order allows the refinement
            # rule to be found in the rule table.)  rotation is the
            # offset into the elements node list required to match the
            # refinement rule to the element's marked edges.
            # signature is the canonical ordering of the marks.
            rotation, signature = findSignature(marks)
            # Create new nodes along the subdivided element edges
            edgenodes = [self.getNewEdgeNodes(oldElement.nodes[i],
                                              oldElement.nodes[(i+1)%oldnnodes],
                                              marks[i], newSkeleton)
                         for i in range(oldnnodes)]
            # Create new elements
            newElements = self.rules[signature].apply(oldElement, rotation,
                                                      edgenodes, newSkeleton)

            # The calls to Skeleton.newElement() made by the
            # refinement rules have created new SkeletonSegments in
            # newSkeleton, but have not set the parentage of those
            # segments.  We have to fix that here.
            for newElement in newElements:
                for segment in newElement.getSegments(newSkeleton):
                    # Only look at each segment once.
                    if not segmentdict.has_key(segment):
                        segmentdict[segment] = 1
                        pseg = findParentSegment(skeleton, newElement, segment,
                                                 edgenodes)
                        if pseg:
                            pseg.add_child(segment)
                            segment.add_parent(pseg)
            if pBar.query_stop():
                pBar.set_failure()
                pBar.set_message("Failed")
                return None
            else:
                pBar.set_progress(1.0*(ii+1)/n)
                pBar.set_message("refining skeleton: %d/%d" % (ii+1, n))

            
        newSkeleton.cleanUp()
        return newSkeleton

    ################
    
    def getNewEdgeNodes(self, node0, node1, ndivs, newSkeleton):
        # Create ndivs new nodes between node0 and node1 in the
        # skeleton.
        if ndivs < 1:
            return []
        key = utils.canonicalorder(node0, node1)
        try:
            nodes = self.newEdgeNodes[key]
            # Since this is the second time we're using this list of
            # nodes, we must be looking at them from the other side,
            # and the nodes should be in the opposite order.
            # Reversing them in place like this would be wrong if the
            # list weren't being used immediately, as it is in
            # Refine.apply()
            nodes.reverse()
        except KeyError:
            nodes = [None]*ndivs
            fixedx = not (node0.movable_x() or node1.movable_x())
            fixedy = not (node0.movable_y() or node1.movable_y())
            p0 = node0.position()
            p1 = node1.position()
            delta = (p1 - p0)/(ndivs+1)
            for i in range(ndivs):
                pt = p0 + (i+1)*delta
                newnode = newSkeleton.newNode(pt.x, pt.y)
                if fixedx:
                    newnode.setMobilityX(0)
                if fixedy:
                    newnode.setMobilityY(0)
                nodes[i] = newnode
            self.newEdgeNodes[key] = nodes
        return nodes

#################################

def findParentSegment(oldSkeleton, element, segment, edgenodes):
    n0, n1 = segment.nodes()            # nodes of the segment
    parents = n0.getParents()
    if parents:
        p0 = parents[-1]
    else:
        p0 = None
        
    parents = n1.getParents()
    if parents:
        p1 = parents[-1]
    else:
        p1 = None    
    oldElement = element.getParents()[-1]
    nnodes = oldElement.nnodes()

    if p0 is not None and p1 is not None:
        pseg = oldSkeleton.findSegment(p0, p1)
        if pseg:
            # Trivial case: the parent nodes define a segment in the
            # parent skeleton.  That segment is this segment's parent.
            return pseg
        # Both nodes have parents, but the parent nodes don't define a
        # segment in the old skeleton.  (This is unlikely to happen
        # during refinement.)  The segment does not have a parent.
        return None

    elif p0 is not None or p1 is not None:
        # Exactly one new endpoint has a parent in the old element.
        # The other endpoint may lie on an old segment.  If that
        # segment's other node is the parent of the new node, then
        # that segment is the parent of this segment.
        if p0 is None:
            parentnode = p1
            childnode = n1
            freechild = n0
        else:
            parentnode = p0
            childnode = n0
            freechild = n1
        which = oldElement.nodes.index(parentnode)
        if freechild in edgenodes[which]:
            next = (which+1)%nnodes
            return oldSkeleton.findSegment(parentnode, oldElement.nodes[next])
        prev = (which - 1 + nnodes) % nnodes
        if freechild in edgenodes[prev]:
            return oldSkeleton.findSegment(parentnode, oldElement.nodes[prev])

    elif p0 is None and p1 is None:
        # Neither new endpoint has a parent.  If the new endpoints are
        # consecutive nodes in the list of new nodes added to a parent
        # element segment, then that parent segment is the parent
        # we're looking for.
        for i in range(nnodes):
            nodelist = edgenodes[i]
            try:
                which0 = nodelist.index(n0)
                which1 = nodelist.index(n1)
            except ValueError:
                pass
            else:
                if abs(which0 - which1) == 1:
                    return oldSkeleton.findSegment(oldElement.nodes[i],
                                                   oldElement.nodes[(i+1)%
                                                                    nnodes])

################################

def findSignature(marks):
    # Given a list of subdivisions of sides (marks), rotate a
    # canonical starting point for the list, so that it can be used as
    # an index into a table of refinement functions.  For a list
    # [x,y,z], the canonical order is that which maximizes the number
    # xyz in base arbitrary_factor.  This will fail if edges are ever
    # divided into more than arbitrary_factor segments in one
    # refinement operation.
    n = len(marks)
    max = -1
    imax = None
    for i in range(n):
        key = marks[i]
        for j in range(1, n):
            key = arbitrary_factor*key + marks[(i+j)%n]
        if key > max:
            max = key
            imax = i
    return imax, tuple([marks[(i+imax)%n] for i in range(n)])


####################

registeredclass.Registration('Refine', skeletonmodifier.SkeletonModifier,
                             Refine,
                             ordering=0,
                             params=[
    parameter.Parameter('targets', checkandmark.CheckAndMark,
              tip='Target elements to be refined'),
    parameter.Parameter('criterion', RefinementCriterion,
                        tip='Refinement criterion'),
    parameter.Parameter('degree', RefinementDegree,
                        tip='Preferred way of subdividing a side')
    ],
                             tip="Subdivide elements.")
