# -*- python -*-
# $RCSfile: skeletonnode.py,v $
# $Revision: 1.32 $
# $Author: langer $
# $Date: 2004/10/07 21:30:24 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov.

from SWIG.common import switchboard
from SWIG.engine import cskeleton
from engine import skeletonselectable
from common import debug
from common import object_id
from common import primitives
from common import utils
import weakref

class SkeletonNode(skeletonselectable.SkeletonSelectable,
                   cskeleton.CSkeletonNode):
    count = 0

    def __init__(self, x, y):
        cskeleton.CSkeletonNode.__init__(self, x, y)
        skeletonselectable.SkeletonSelectable.__init__(self)
        # Get unique index.
        self.index = SkeletonNode.count
        SkeletonNode.count += 1
        self._elements = []
        self.ID = object_id.ObjectID()
        self.meshindex = None
    def __repr__(self):
##        p = self.position()
##         return "SkeletonNode(%f, %f, %d)" % (p.x, p.y, self.index)
        return "SkeletonNode(%d)" % self.index

    def __eq__(self, other):
        return isinstance(other, SkeletonNode) and self.index == other.index

    def __hash__(self):
        return self.index

    def __cmp__(self, other):
        if not isinstance(other, SkeletonNode):
            return -1
        else:
            return cmp(self.index, other.index)

    def __lt__(self, other):
        if not isinstance(other, SkeletonNode):
            return -1
        else:
            return self.index < other.index
    
    def repr_position(self):
        return self.position()

    def getIndex(self):
        return self.index

    # There may be several meshes per skeleton, but the corresponding
    # node in all of those meshes will have this same index.
    def setMeshIndex(self, value):
        self.meshindex = value

    def getMeshIndex(self):
        return self.meshindex
    
    def destroy(self, skeleton):
        skeleton.removeNode(self)
        self.disconnect()
        del self._elements

    def active(self, skeleton):
        return skeleton.MS.activePoint(self.position())

    def illegal(self):
        # A node is illegal if any of its elements are illegal.  This
        # is used when testing possible node motions, so it's assumed
        # that the node being tested is the one causing the elements
        # to be illegal.
        for element in self._elements:
            if element.illegal():
                return 1
        return 0

    def addElement(self, element):
        self._elements.append(element)

    def removeElement(self, skeleton, element):
        self._elements.remove(element)
        if not self._elements:
            self.destroy(skeleton)
        
    def neighborElements(self):
        return self._elements

    def neighborEnergy(self, skeleton, alpha):
        e = 0.0
        for element in self._elements:
            e += element.energyTotal(skeleton, alpha)
        return e

    def neighborNodes(self,skeleton):
        neighborDict = {}
        for e in self._elements:
            for nd in e.nodes:
                if nd is not self:
                    if skeleton.findSegment(nd, self) is not None:
                        neighborDict[nd] = 1
        return neighborDict.keys()

    # "Segments for which I am an endpoint."
    def localSegments(self,skeleton):
        result_dict = {}
        for e in self._elements:
            for nd in e.nodes:
                if nd is not self:
                    s = skeleton.findSegment(nd, self)
                    if s:
                        result_dict[s]=1
        return result_dict.keys()
        
    def moveBack(self):
        t = self.nodemoved
        cskeleton.CSkeletonNode.moveBack(self)
        for el in self._elements:
            el.revert()

    def moveToNode(self, target):
        self.moveTo(target.position())
        self.copyMobility(target)
        for el in target.neighborElements():
            if el not in self._elements:
                self.addElement(el)
        
    def new_child(self):
        p = self.position()
        new = SkeletonNode(p.x, p.y)
        new.copyMobility(self)
        return new

    ####################

    # pin and unpin are analogous to select and deselect in the
    # SkeletonSelectable class.  They pin and unpin this node and its
    # children and parents.  The clist and plist args are lists of
    # trackers corresponding to the children and parents in the
    # SkeletonContext's stack of skeletons.  Some of those Skeletons
    # may be DeputySkeletons, which share nodes instead of having
    # parent/child nodes.  For those Skeletons the tracker is a
    # DeputyPinnedNodeTracker instead of a PinnedNodeTracker.

    def pin(self, clist, plist):
        self.setPinned(1)
        here = self.position()
        self.pinDown(clist, here)
        self.pinUp(plist, here)
    def pinDown(self, clist, here):
        if clist[0].nodePosition(self) == here:
            clist[0].add(self)
            if len(clist) > 1:
                clist[1].pinDown(self, clist[1:], here)
    def pinUp(self, plist, here):
        if plist[0].nodePosition(self) == here:
            plist[0].add(self)
            if len(plist) > 1:
                plist[0].pinUp(self, plist[1:], here)
            
    def unpin(self, clist, plist):
        self.setPinned(0)
        here = self.position()
        self.unpinDown(clist, here)
        self.unpinUp(plist, here)
    def unpinDown(self, clist, here):
        if clist[0].nodePosition(self) == here:
            clist[0].remove(self)
            if len(clist) > 1:
                clist[1].unpinDown(self, clist[1:], here)
    def unpinUp(self, plist, here):
        if plist[0].nodePosition(self) == here:
            plist[0].remove(self)
            if len(plist) > 1:
                plist[0].unpinUp(self, plist[1:], here)
        
#######################################################

class HashedNodes:
    def __init__(self, size, skelsize):
        self.size = size
        n = size[0]*size[1]
        self.scale = (1.0*self.size[0]/skelsize[0],
                      1.0*self.size[1]/skelsize[1])
        self.data = [[] for i in range(n)]

    def getSize(self):
        return self.size

    def getScale(self):
        return self.scale

    def properTile(self, point):
        ix = int( point.x*self.scale[0] )
        iy = int( point.y*self.scale[1] )
        # When right or top skeleton edge was clicked ....
        if ix==self.size[0]: ix -= 1
        if iy==self.size[1]: iy -= 1
        return primitives.iPoint(ix, iy)

    def iterator(self):
        return self.data

    def __setitem__(self, where, val):
        self.data[self.size[0]*where.y + where.x] = val

    def __getitem__(self, where):
        try:
            if (where.x>=0 and where.y>=0):
                return self.data[self.size[0]*where.y + where.x]
            else:
                return []
        except IndexError:
            return []

    def hash(self, skeleton):
        # Putting nodes in corresponding tiles
        for node in skeleton.nodes:
            where = self.properTile(node.position())
            self[where].append(node)

    def nearestNode(self, point):
        where = self.properTile(point)

        count = 0  # No. of tile-bands containing nodes
        iter = 1   # No. of tile-bands having been looked at
        nodes = self[where][:]
        while count < 2:   
            # Initial check
            if iter==1:
                if nodes: count += 1
            # Additional band of tiles
            addition = self.nextBand(
                ll=primitives.iPoint(where.x-iter, where.y-iter),
                ur=primitives.iPoint(where.x+iter, where.y+iter))
            if addition:
                nodes += addition
                count += 1
            else:
                if count==1:
                    break  # No need to look for another band
            # Ready for the next round
            iter += 1

        # Find the nearest node among collected nodes
        nearest = nodes[0]
        mindist = (nearest.position() - point)**2
        for node in nodes[1:]:
            dd = (node.position() - point)**2
            if dd < mindist:
                mindist = dd
                nearest = node
        return nearest

    def nextBand(self, ll=None, ur=None):
        nodes = []
        # Bottom
        j = ll.y
        for i in range(ll.x, ur.x+1):
            nodes += self[primitives.iPoint(i,j)][:]
        # Top
        j = ur.y
        for i in range(ll.x, ur.x+1):
            nodes += self[primitives.iPoint(i,j)][:]
        # Left
        i = ll.x
        for j in range(ll.y+1, ur.y):
            nodes += self[primitives.iPoint(i,j)][:]
        # Right
        i = ur.x
        for j in range(ll.y+1, ur.y):
            nodes += self[primitives.iPoint(i,j)][:]
        return nodes

###############################################################
    
# The set of pinned nodes acts much like a set of selected objects, so
# it derives a lot of its code from skeletonselectable.  See comments
# in SelectionSetBase and SelectionSet in skeletonselectable.py.

class PinnedNodeSet(skeletonselectable.SelectionSetBase):
    def __init__(self, *args, **kwargs):
        skeletonselectable.SelectionSetBase.__init__(self, *args, **kwargs)

    def clearskeletons(self):
        for tracker in self.selected.values():
            for n in tracker.data:
                n.setPinned(0)

    def implied_select(self, oldskel, newskel):
        # Called from SelectionBase.whoChanged0() when a new
        # Skeleton is pushed.  Creates a tracker for the new skeleton
        # with the same pinned nodes as the old skeleton, if those
        # nodes are at the old positions.
        tracker = newskel.newPinnedNodeTracker()
        self.selected[newskel] = tracker
        if oldskel is not newskel:
            oldtracker = self.selected[oldskel]
            tracker.implied_pin(oldtracker)

    def replace(self, oldskel, newskel):
        # Called just after a SkeletonContext undo or redo by
        # SkeletonContext.undoHookFn and SkeletonContext.redoHookFn.
        oldtracker = self.selected[oldskel]
        oldtracker.clearskeleton()
        newtracker = self.selected[newskel]
        newtracker.write()

    def clone(self):
        bozo = self.__class__(self.skeletoncontext)
        for skel, tracker in self.selected.items():
            trclone = tracker.clone()
            bozo.selected[skel] = trclone
        return bozo

class PinnedNodeTracker(skeletonselectable.SelectionTrackerBase):
    def __init__(self, skeleton):
        self.skeleton = weakref.ref(skeleton)
        skeletonselectable.SelectionTrackerBase.__init__(self)

    def clone(self):
        bozo = self.__class__(self.skeleton())
        bozo.data += self.data[:]
        return bozo

    def nodePosition(self, node):
        skel = self.skeleton()
        return skel.nodePosition(node)
        
    def clear(self):
        for n in self.data:
            n.setPinned(0)
        self.data = utils.UniqueList()

    def write(self):
        for n in self.data:
            n.setPinned(1)
            
    def clearskeleton(self):
        for n in self.data:
            n.setPinned(0)

    def implied_pin(self, oldtracker):
        for n in oldtracker.get():
            there = oldtracker.nodePosition(n)
            for child in n.getChildren():
                here = self.nodePosition(child)
                if here == there:
                    self.add(child)
                    
    def pinDown(self, pnode, clist, where):
        for c in pnode.getChildren():
            c.pinDown(clist, where)
    def pinUp(self, cnode, plist, where):
        for p in cnode.getParents():
            p.pinUp(plist, where)

    def unpinDown(self, pnode, clist, where):
        for c in pnode.getChildren():
            c.unpinDown(clist, where)
    def unpinUp(self, cnode, plist, where):
        for p in cnode.getParents():
            p.unpinUp(plist, where)
            
    def __repr__(self):
        return "PinnedNodeTracker"


class PinnedNodeSelection(skeletonselectable.SelectionBase):
    def __init__(self, skeletoncontext):
        skeletonselectable.SelectionBase.__init__(self, skeletoncontext)
        self.stack.push(PinnedNodeSet(self.skeletoncontext))

    def replace(self, oldskel, newskel):
        # Called just after a SkeletonContext undo or redo by
        # SkeletonContext.undoHookFn and SkeletonContext.redoHookFn.
        self.stack.current().replace(oldskel, newskel)

    def signal(self):
        # signal is called after each pinning operation (by the menu
        # commands in the toolbox).  It's also called by the generic
        # SkeletonSelection mechanism, which is why the menu commands
        # don't just simply issue the "new pinned nodes" message
        # themselves.  "new pinned nodes" is caught by all the
        # toolboxes, which relay the message to their graphics
        # windows.
        switchboard.notify("new pinned nodes", self)

    def pin(self, nodelist):
        (clist, plist) = self.trackerlist()
        for n in nodelist:
            n.pin(clist, plist)
        self.timestamp.increment()

    def unpin(self, nodelist):
        (clist, plist) = self.trackerlist()
        for n in nodelist:
            n.unpin(clist, plist)
        self.timestamp.increment()

    def toggle(self, nodelist):
        (clist, plist) = self.trackerlist()
        for n in nodelist:
            if n.pinned():
                n.unpin(clist, plist)
            else:
                n.pin(clist, plist)
        self.timestamp.increment()

    def pinPoint(self, x, y):           # oxford
        skel = self.skeletoncontext.getObject()
        node = skel.nearestNode(primitives.Point(x,y))
        self.pin([node])

    def unpinPoint(self, x, y):
        skel = self.skeletoncontext.getObject()
        node = skel.nearestNode(primitives.Point(x,y))
        self.unpin([node])
                
    def togglepinPoint(self, x, y):
        skel = self.skeletoncontext.getObject()
        node = skel.nearestNode(primitives.Point(x,y))
        self.toggle([node])
